{{
/**
 *******************************************************************************************************
 * @file jvmdoc.hh
 *
 * @note Jazzed_JVM_Demo_008.spin
 *      Added public front end API.
 *      Added debug menu for API testing and JVM debug.
 *      Issues: menu command i causes r & c work only once per boot.
 *      WorkAround: avoid using i command :).
 *
 * @note Jazzed_JVM_Demo_006.spin.
 *      Added handling for InvokeVirtual.
 *      Added Peter's native method code.
 *      Hello World program works :).
 *
 * @note Jazzed_JVM_Demo_004.spin.
 *      Added support for LDC and siblings.
 *      Added support for GETFIELD/PUTFIELD.
 *      Added support for INSTANCEOF.
 *      Added support for NEW, NEWARRAY, and ANEWARRAY.
 *      Added code to InvokeProgram to get us further along.
 *
 * @note Jazzed_JVM_Demo_002.spin.
 *      Merged debug support with AiChip_JVM_Demo_002.spin.
 *
}}

CON
  

DAT ' jazzDebug , debugMsDelay < 0 means stop for debug
  jazzDebug             byte 0
  callDebug             byte 0
  natDebug              byte 0
  debugging             byte 0 ' Set true if debugging the JVM
  debugMsDelay          long 0
  gAbort                long 0 ' allows user to abort JVM run if set
                               
CON
{  
As supplied the jem.out is ...

'        import stamp.*;
'        public class testJVM
'        {
'          static int iStatic;
'          static int incr( int a )
'          {
'            return a + 1;
'          }
'          static void main()
'          {
'            iStatic = 0;
'            while (true)
'            {
'              iStatic = java.lang.Math.min( incr(iStatic), 10);
'              stamp.core.CPU.delay(iStatic);
'            }
'          }
'        }

Progress so far ...

        DONE - Get 'main' running by itself using local variables
        DONE - Get 'main' calling other class native methods
        DONE - Get 'main' calling same class static method
        DONE - Get 'main' calling other class static methods
        DONE - Get 'main' called from program initialisation
        DONE - Create simple heap handler
        DONE - Handle array creation
        DONE - Handle class static variables for 'testJVM'
        DONE - Get main running with class static variables for 'testJVM'

        TODO - Add Native Method handling code
        Added Peter's native library, handler, and parameter passing ... jazzed
        
Not Implemented yet ...

        Jem_INVOKEVIRTUAL       jazzed
        Jem_INVOKEINTERFACE
        Jem_ATHROW       
        Jem_CHECKCAST
        Jem_INSTANCEOF          jazzed
        Jem_WIDE
        
        Jem_GETFIELD            jazzed 
        Jem_PUTFIELD            jazzed 
        Jem_NEW                 jazzed
        
        Jem_ANEWARRAY           jazzed
        
        Jem_LDC                 jazzed
        Jem_LDC_W
        Jem_LDC2_W

        Jem_TABLESWITCH
        Jem_LOOKUPSWITCH

        PopObjectRef            Need for IF_ACMPEQ/NE
      
Not completely implemented ...

        InvokeProgram           Just grabs an array for static fields ...  jazzed
        Jem_NEWARRAY            What does the operand mean ? Always $00 !  jazzed

}
CON
{

PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_Jvm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}

OBJ
  jem   : "Jazzed_JVM_JemCodes_001"                     ' JEM Bytecode to execute ( jem.bin )
  jcc   : "Jazzed_JVM_CrashCodes_001"                   ' JVM Crash Codes
  'jni   : "Jazzed_jvmNative"
  
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN              = 16 '12                              ' Base TV Pin

CON

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Configuration                                                                               |
' `-----------------------------------------------------------------------------------------------------'

  INT32               = false                           ' Set true if an int is 32-bit, otherwise 16-bit
  LNG64               = false                           ' Set true if a long is 64-bit, otherwise 32-bit
  CHR16               = true                            ' Set true if a char is 16-bit, otherwise 8-bit

  BIGARRAYS           = false                           ' Set true if allowed arrays larger than 64KB
  
  STACK_SIZE          = 1024                            ' Entries allowed on JVM runtime execution stack

  HEAP_BASE           = $7E00                           ' Address of first free byte above top of heap

CON
  
  INTWORDS            = 1 - INT32                       ' Number of words in an int ( INT32=2, INT16=1 )
  INTBYTES            = INTWORDS * 2                    ' Number of bytes in an int ( INT32=4, INT16=2 )

  LNGWORDS            = ( 1 - LNG64 ) * 2               ' Number of words in a long ( LNG64=4, LNG32=2 )
  LNGBYTES            = LNGWORDS * 2                    ' Number of bytes in a long ( LNG64=8, LNG32=4 )

  CHRWORDS            = ( 1 - CHR16 ) >> 1              ' Number of words in a char ( CHR16=1, CHR8=0  )
  CHRBYTES            = 1 - CHR16                       ' Number of bytes in a char ( CHR16=2, CHR8=1  )

  MODE_SPIN_ONLY      = %010                            ' How the JVM opcode processing can be handled
  MODE_SPIN_THEN_PASM = %011
  MODE_PASM_ONLY      = %100
  MODE_PASM_THEN_SPIN = %110

  
  HEAP_HEAD_TYPE      = 0                               ' Heap chunk Type entry offset
  HEAP_HEAD_BLEN      = 1                               ' Heap chunk ByteLen entry offset
  HEAP_HEAD_DATA      = 2                               ' Heap chunk Data entry offset

  HEAP_WORD           = 1
  HEAP_LONG           = 2

  BREAKPOINT_SIZE     = 10

CON 'jvm instruction termination codes

  #-3, jvmCrashed, jvmAborted, jvmBreakPoint, jvmExitOk, jvmPasmModeDisabled, jvmSpinModeDisabled  

DAT

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Runtime Data Structures                                                                     |
' `-----------------------------------------------------------------------------------------------------'

  ' The JVM runtime stack is adjusted for the base
  ' size of int(16) or int(32). Only half as many
  ' longs will be used for int(16) as for int(32).
  ' Keep this as long so it works with INT32.
  
  'long stack[ STACK_SIZE >> (INT32+1) ]                 ' Stack for JVM runtime execution
  stack long 0 [ STACK_SIZE >> (INT32+1) ]               ' Stack for JVM runtime execution

VAR

  ' The following ( 'pc' to 'sp' ) must all be longs
  ' and must be contiguous. Add new stack frame items
  ' within this list, not before 'pc' nor after 'sp'
  ' or the stack frames will not work properly.
  
  long pc                                               ' Current JEM bytecode Pc ( absilute hub address )
  long bp                                               ' Current base of stack ( firat param / local )
  long mb                                               ' Method base - first executable in method
  long fp                                               ' Current frame pointer
  long ct                                               ' Constant table address
  long et                                               ' Exception table address
  long vp                                               ' Variable base ( static fields )
  long sp                                               ' Top of stack pointer ( last pushed )


  long bytecodeStart                                    ' Absolute Hub address of first bytecode
  long jemFileCountAdjust                               ' How to adjust Jem File counts to fit reality

  long heap                                             ' Ptr to last chunk created on heap
  
  long mode                                             ' Type of JVM opcode processing
  long allowNotImplementedNatives                       ' Set true if unimplemented Native Methods ignored

  long stackAltered                                     ' True if stack has altered flag
  long opcode                                           ' Current opcode

  long gBytecodeLen                                     ' length of bytecode (len field * 256)
  long gBytecodeMain                                    ' bytecode file's address for main
  long gCallMainMethodAddr                              ' main method addr to call
  long gClassId                                         ' working object class id
  long gMethodId                                        ' working method id
  long gObjectClassId                                   ' java.lang.Object class id
  long gConstTable                                      ' constant table pointer

  word gGetOverBreak                                    ' skip break once if set during isBreakPoint
  word gBreakpointHead                                  ' breakpoint array queue head
  word gBreakpointArray[BREAKPOINT_SIZE]                ' breakpoint array

  long gNativeCode                                      ' if => 0 after dobytecode, let main handle it
   
  'long cog
  'long cogstack[600]


CON
  _clkmode = xtal1+pll16x
  _xinfreq = 5_000_000

OBJ
  'tv : "Jazzed_VGA_Text" ' serial debug
  tv : "Jazzed_SIO_Text"
  'tv : "Jazzed_SIO_Text_Stub"
  
PUB main
  start
  
PUB start
{{
/**
 * This is the startup point for module using the JVM
 * A new COG is not initialized for this today.
 */
__ void start();
}}
  'tv.start(16)
  'tv.start(31, 30, 0, 28800)

  mode := MODE_SPIN_ONLY
  gBreakpointHead := 0
  heap := HEAP_BASE
  gAbort := false
  allowNotImplementedNatives := true
  'cog := cognew(mainrun, @cogstack) + 1
  
PUB stop
{{
/**
 * This is a compatibility stop function.
 * A new COG is not initialized for the module, so nothing is stopped.
 */
__ void stop();
}}
  'if cog
  '   cogstop(cog~ -  1)

'PRI mainrun

PUB jvmSetup | ii
{{
/**
 * Run the JVM init procedure without running main. This resets the JVM engine.
 */
__ void jvmInit();
}}

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise JVM                                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  sp := @stack - INTBYTES
  bp := sp
  mb := 0
  fp := 0
  pc := 0
  if heap <> 0
    repeat ii from heap to HEAP_BASE
      byte[ii] := 0  
  heap := HEAP_BASE
  
  gAbort := false
  InvokeProgram( bytecodeStart, false )
  
PUB jvmInit
{{
/**
 * Run the JVM init procedure. This resets the JVM engine and starts main.
 */
__ void jvmInit();
}}
  jvmSetup
  CallMethod(gCallMainMethodAddr, false)  
  'InvokeProgram( bytecodeStart, true )

PUB dumpRegs
  tv.str(string($d,"PC "))
  tv.hex(getJvmPC, 4)
  tv.str(string(" SP.SB.FP.MP.HP.Obj "))
  tv.hex(getJvmSP, 4)
  tv.Out(".")
  tv.hex(getJvmSB, 4)
  tv.Out(".")
  tv.hex(getJvmFP, 4)
  tv.Out(".")
  tv.hex(getJvmMP, 4)
  tv.Out(".")
  tv.hex(getJvmHP, 4)
  tv.Out(".")
  tv.hex(gClassId, 4)
  tv.Out(" ")
  tv.str(string(" JPC "))
  tv.hex( getJvmMP-bytecodeStart, 4)
  tv.Out( ":" )
  tv.Hex( pc - bytecodeStart, 4 )
{
  tv.Out( " " )
  tv.Dec( pc - mb )
  if (pc - mb ) => 0 and ( pc - mb ) < 10
    tv.Out( " " )
  if ( pc - mb ) < 100
    tv.Out( " " )
}

PUB dumpOpcode(opc, opname)   
  tv.Str(string("OP "))
  tv.Hex(opc, 2)
  'tv.Str(string(" - "))
  tv.Out(" ")
  tv.Str(opname)
    
PUB JvmSpinDumpBreak(opc, opname) | ch
{{
/**
 * Runs the JVM until a breakpoint is hit or if singlestep is non-zero runs one instruction.
 * @returns 0 if function terminates normally.
 */
__ void JvmSpinDumpBreak(opcode, opname);
}}
  dumpRegs
  dumpOpcode(opc,opname)
 

PUB JvmSpinDebug(singlestep, verbose, opname)
{{
/**
 * Runs the JVM until a breakpoint is hit or if singlestep is non-zero runs one instruction.
 * @returns 0 if function terminates normally.
 */
__ int jvmSpinDebug(singlestep, verbose);
}}
' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  repeat
  
    gNativeCode := -1
    result := gNativeCode            

    if verbose
      dumpRegs

    ' with native processing outside of JVM, we have to do this outside too.
    '
    'if isBreakpoint(pc)
    '  return result

    opcode := byte[ pc++ ]
     
    if verbose
      dumpOpcode(opcode,opname)
     
    \JvmSpin

    result := gNativeCode    

    if singlestep
      return result

PUB doBytecode

  gNativeCode := -1
  opcode := byte[ pc++ ]
  JvmSpin

  result := gNativeCode
{  
PRI JvmSpinLoop | rcnt

' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  repeat

    if gAbort
      gAbort := false
      return 0
  
      opcode := byte[ pc++ ]
      JvmSpin
      
    handleNative(gNativeCode)
        
  Crashed(jcc#JvmSpin_terminated)

PRI handleNative(code) | stk, ii, rcnt, val

  ' native methods are positive integers
  if code < 0
    return

  rcnt := Pop

  repeat while (rcnt-- > 0)
    val := GetVar(rcnt)
    result := GetHeapValuePtr(val,0)
    if result <> 0
      val := result
    jni.push(val)

  result := jni.do_native(code, @rcnt)

  if rcnt > 0
    Push(result)
}   
    
PRI JvmSpin | arrayPtr, arrayLen, arrayIndex, value

  if gAbort <> 0
    return jvmAborted
    
  if mode & MODE_PASM_ONLY
    if JvmPasm
       return jvmPasmModeDisabled
    else
      if mode == MODE_PASM_ONLY
        return jvmSpinModeDisabled
        'Crashed(jcc#PASM_cant_handle_opcode)

  if mode & MODE_SPIN_ONLY
              
    if opcode < jem#Jem_INSTANCEOF
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     NOP                                                                                 $00 |
        ' |     Meaning: Do nothing.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NOP:

        ' .---------------------------------------------------------------------------------------------.
        ' |     ICONST_M1                                                                           $01 |
        ' |     Meaning: Push the integer constant -1.                                                  |
        ' |     Description: This bytecode pushes the integer constant -1 onto the stack.               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ICONST_0                                                                        $02-$07 |
        ' |     Meaning: Push the integer constant <C>.                                                 |
        ' |     Description: This bytecode pushes an integer constant onto the stack. There             |
        ' |     are six associated bytecodes, one for each of the integers 0-5: iconst_0,               |
        ' |     iconst_1, iconst_2, iconst_3, iconst_4, and iconst_5.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ICONST_M1, jem#Jem_ICONST_0, jem#Jem_ICONST_1, jem#Jem_ICONST_2, jem#Jem_ICONST_3:

          Push( opcode - jem#Jem_ICONST_0 )

        jem#Jem_ICONST_4, jem#Jem_ICONST_5:

          Push( opcode - jem#Jem_ICONST_4 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BIPUSH                                                                              $08 |
        ' |     Meaning: Push a 1-byte signed integer.                                                  |
        ' |     Description: The first byte following the opcode, byte1, is interpreted as a            |
        ' |     signed 8-bit value. This value is expanded to a 32-bit integer and pushed onto          |
        ' |     the operand stack.                                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BIPUSH:

          Push( Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     SIPUSH                                                                              $09 |
        ' |     Meaning: Push a 2-byte signed integer.                                                  |
        ' |     Description: The first and second byte following the opcode, byte1 and byte2,           |
        ' |     are assembled into a signed 16-bit value. This value is expanded to a 32-bit            |
        ' |     integer and pushed onto the operand stack.                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_SIPUSH:

          Push( Operand_SignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ILOAD                                                                               $0A |
        ' |     Meaning: Load integer from local variable.                                              |
        ' |     Description: The value of the local variable in the current Java frame is               |
        ' |     pushed onto the operand stack. This value must be of type integer.                      |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ILOAD_0                                                                             $0B |
        ' |     ILOAD_1                                                                             $0C |
        ' |     ILOAD_2                                                                             $0D |
        ' |     ILOAD_3                                                                             $0E |
        ' |     Meaning: Load integer from local variable with specific index.                          |
        ' |     Description: The value of the local variable <l> in the current Java frame is           |
        ' |     pushed onto the operand stack.                                                          |
        ' |     This value must be of type integer. There are four of these bytecodes, one for          |
        ' |     each of the integers 0-3: iload_0, iload_1, iload_2, and iload_3.                       |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ILOAD:

           Push( GetVar( Operand_UnsignedByte ) )

        jem#Jem_ILOAD_0, jem#Jem_ILOAD_1, jem#Jem_ILOAD_2, jem#Jem_ILOAD_3:

           Push( GetVar( opcode - jem#Jem_ILOAD_0 ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ISTORE                                                                              $0F |
        ' |     Meaning: Store integer into local variable.                                             |
        ' |     Description: Local variable in the current Java frame is set to a value. This           |
        ' |     value must be an integer.                                                               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ISTORE_0                                                                            $10 |
        ' |     ISTORE_1                                                                            $11 |
        ' |     ISTORE_2                                                                            $12 |
        ' |     ISTORE_3                                                                            $13 |
        ' |     Meaning: Store integer into local variable with specific index.                         |
        ' |     Description: Local variable <l> in the current Java frame is set to a value.            |
        ' |     This value must be an integer. There are four of these bytecodes, one for each          |
        ' |     of the integers 0-3: istore_0, istore_1, istore_2, and istore_3.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ISTORE:

           PutVar( Operand_UnsignedByte, Pop )

        jem#Jem_ISTORE_0, jem#Jem_ISTORE_1, jem#Jem_ISTORE_2, jem#Jem_ISTORE_3:

           PutVar( opcode - jem#Jem_ISTORE_0, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IFxx                                                                                $14 |
        ' |     Meaning: Branch if condition                                                            |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IFEQ: BranchIf( Pop == 0 )
        jem#Jem_IFNE: BranchIf( Pop <> 0 )
        jem#Jem_IFLT: BranchIf( Pop <  0 )
        jem#Jem_IFLE: BranchIf( Pop =< 0 )
        jem#Jem_IFGT: BranchIf( Pop >  0 )
        jem#Jem_IFGE: BranchIf( Pop => 0 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ICMPxx                                                                           $1A |
        ' |     Meaning: Branch if compare                                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ICMPEQ: BranchIf( PopL == PopR )
        jem#Jem_IF_ICMPNE: BranchIf( PopL <> PopR )
        jem#Jem_IF_ICMPLT: BranchIf( PopL <  PopR )
        jem#Jem_IF_ICMPGE: BranchIf( PopL => PopR )
        jem#Jem_IF_ICMPGT: BranchIf( PopL >  PopR )
        jem#Jem_IF_ICMPLE: BranchIf( PopL =< PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ACMPxx                                                                           $20 |
        ' |     Meaning: Branch if object references equal.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ACMPEQ: BranchIf( PopObjectRef == PopObjectRef )
        jem#Jem_IF_ACMPNE: BranchIf( PopObjectRef <> PopObjectRef )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IRETURN                                                                             $22 |
        ' |     Meaning: Return integer from function.                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IRETURN: ReturnFromMethodWithValue
        jem#Jem_RETURN:  ReturnFromMethod

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEVIRTUAL                                                                       $24 |
        ' |     Meaning: Invoke method based on runtime type.                                           |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEVIRTUAL:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKESTATIC                                                                        $25 |
        ' |     Meaning: Invoke a class (static) method.                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKESTATIC:

          InvokeStatic( Operand_BytecodeAddress )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETSTATIC                                                                           $26 |
        ' |     Meaning: Get 16bit static field from class.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETSTATIC:

          Push( Peek( vp, Operand_UnsignedWord >> jemFileCountAdjust ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTSTATIC                                                                           $27 |
        ' |     Meaning: Set 16bit static field in class.                                               |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTSTATIC:

          Poke( vp, Operand_UnsignedWord >> jemFileCountAdjust, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     POP                                                                                 $28 |
        ' |     Meaning: Stack manipulation                                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_POP:     Pop
        jem#Jem_POP2:    Pop2
        jem#Jem_DUP:     Dup
        jem#Jem_DUP_X1:  Dup_X1
        jem#Jem_DUP_X2:  Dup_X2
        jem#Jem_DUP2:    Dup2
        jem#Jem_DUP2_X1: Dup2_X1
        jem#Jem_DUP2_X2: Dup2_X2
        jem#Jem_SWAP:    Swap

        ' .---------------------------------------------------------------------------------------------.
        ' |     IADD, ISUB, INEG, IOR, ISHL, ISHR, IUSHR, IAND, IXOR                                $31 |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IADD:  Push( PopL +  PopR )
        jem#Jem_ISUB:  Push( PopL -  PopR )
        jem#Jem_INEG:  Push(      -  Pop  )
        jem#Jem_IOR:   Push( PopL |  PopR )
        jem#Jem_ISHL:  Push( PopL << PopR )
        jem#Jem_ISHR:  Push( PopL -> PopR )
        jem#Jem_IUSHR: Push( PopL >> PopR )
        jem#Jem_IAND:  Push( PopL &  PopR )
        jem#Jem_IXOR:  Push( PopL ^  PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     I2B                                                                                 $3A |
        ' |     Meaning: Integer to signed byte.                                                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_I2B:

          if Tos <> SignExtendByte( Tos )
            Crashed(jcc#Signed_Int_too_big)
             
        ' .---------------------------------------------------------------------------------------------.
        ' |     GOTO                                                                                $3B |
        ' |     Meaning: Branch.                                                                        |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2.                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GOTO:

          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEINTERFACE                                                                     $3C |
        ' |     Meaning: Invoke interface method.                                                       |
        ' |     Note                                                                                    |
        ' |     Unlike invokevirtual and invokespecial, the method block does not indicate the          |
        ' |     number of available arguments. This number is taken from the bytecode.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEINTERFACE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     ARRAYLENGTH                                                                         $3D |
        ' |     Meaning: Get length of array.                                                           |
        ' |     Description: The length of the array is determined and replaces aref on the             |
        ' |     top of the stack. The aref must be a reference to an array object.                      |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if the aref is null.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ARRAYLENGTH:

          PopArrayRef( jem#T_ANY, @arrayPtr, @arrayLen )
          Push( arrayLen)
          
        ' .---------------------------------------------------------------------------------------------.
        ' |     ATHROW                                                                              $3E |
        ' |     Meaning: Throw exception or error.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ATHROW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     CHECKCAST                                                                           $3F |
        ' |     Meaning: Make sure object is of given type.                                             |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of another class. A null object reference can be cast to         |
        ' |     any class. Otherwise, the referenced object must be an instance of the expected         |
        ' |     class or one of its superclasses. byte1 and byte2 are used to construct an              |
        ' |     index into the constant pool of the current class, which is presumed to be a            |
        ' |     class name that can be resolved to a class pointer.                                     |
        ' |     Note                                                                                    |
        ' |     A ClassCastException is thrown if the referenced object cannot be cast to the           |
        ' |     expected class.                                                                         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_CHECKCAST:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

    else
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     INSTANCEOF                                                                          $40 |
        ' |     Meaning: Determine if an object is of given type and return result.                     |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of the expected class. This instruction will overwrite           |
        ' |     the object reference with 1 if the object is an instance of the expected class          |
        ' |     or one of its superclasses. Otherwise, the object reference is overwritten by 0.        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INSTANCEOF:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     WIDE                                                                                $41 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_WIDE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IINC                                                                                $42 |
        ' |     Meaning: Increment local variable by constant.                                          |
        ' |     Description: Local variable at byte1 in the current Java frame must contain an          |
        ' |     integer. Its value is incremented by the value byte2, where byte2 is treated as         |
        ' |     a signed 8-bit quantity.                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IINC:

          IncVar( Operand_UnSignedByte, Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETFIELD                                                                            $43 |
        ' |     Meaning: Get 16bit field from object.                                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTFIELD                                                                            $44 |
        ' |     Meaning: Set 16bit field in object.                                                     |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEW                                                                                 $45 |
        ' |     Meaning: Create new object.                                                             |
        ' |     Description: byte1 and byte2 are used to construct an index into the constant           |
        ' |     pool of the current class, which must be a class name that can be resolved to a         |
        ' |     class pointer. A new instance of that class is then created, and a reference to         |
        ' |     the object is pushed on the stack.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEWARRAY                                                                            $46 |
        ' |     Meaning: Allocate new array.                                                            |
        ' |     Description: A new array of a specific array type, capable of holding size              |
        ' |     elements, is allocated. The result is a reference to the new object. Allocation         |
        ' |     of an array large enough to contain size items of the specific array type is            |
        ' |     attempted and all elements of the array are initialized to 0.                           |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. The result is stored with an internal code that indicates the type of          |
        ' |     array to allocate. Possible values for the type of array are as follows:                |
        ' |     T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),                |
        ' |     T_INT(10), and T_LONG(11).                                                              |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEWARRAY:

          NotImplemented
{                    
          arrayLen := Pop

          Operand_UnsignedByte ' ***** HACK **** This is $00 so what's it mean ?
          
          Push( NewArrayRef( arrayType, arrayLen ) )
}
        ' .---------------------------------------------------------------------------------------------.
        ' |     ANEWARRAY                                                                           $47 |
        ' |     Meaning: Allocate new array of objects.                                                 |
        ' |     Description: A new array of the indicated class type and capable of holding             |
        ' |     size elements is allocated. The result is a reference to the new object.                |
        ' |     Allocation of an array large enough to contain size elements of the given class         |
        ' |     type is attempted and all elements of the array are initialized to null.                |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. byte1 and byte2 are used to construct an index into the constant pool          |
        ' |     of the current class. When the item at that index is resolved, the resulting            |
        ' |     entry must be a class.                                                                  |
        ' |     The anewarray instruction is used to create a single-dimension array.                   |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ANEWARRAY:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IALOAD                                                                              $48 |
        ' |     Meaning: Load integer from array.                                                       |
        ' |     Description: The integer value at the array index is retrieved and pushed onto          |
        ' |     the top of the stack. The aref must be a reference to an array of integers;             |
        ' |     likewise, the index into the array must be an integer.                                  |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if aref is null. An                                    |
        ' |     ArrayIndexOutOfBoundsException is thrown if the array index is not within the           |
        ' |     bounds of the array.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed(jcc#Array_Index_oof_bound) 

          if INT32
            push( long[ arrayPtr + ( arrayIndex << 2 ) ] )
          else
            Push( SignExtendWord( word[ arrayPtr + ( arrayIndex << 1 ) ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IASTORE                                                                             $49 |
        ' |     Meaning: Store into integer array.                                                      |
        ' |     Description: An integer value is popped off the stack and stored in the array           |
        ' |     at the index. The aref must be a reference to an array of integer values, and           |
        ' |     the index into the array must be an integer as well.                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed(jcc#Array_Index_oof_bound) 
             
          if INT32
            long[ arrayPtr + ( arrayIndex << 2 ) ] := value
          else
            word[ arrayPtr + ( arrayIndex << 1 ) ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     BALOAD                                                                              $4A |
        ' |     Meaning: Load signed byte from array.                                                   |
        ' |     Description: The signed byte value at the array index is retrieved, expanded            |
        ' |     to an integer, and pushed onto the top of the stack. The aref must be a                 |
        ' |     reference to an array of signed byte values, and the index into the array must          |
        ' |     be an integer.                                                                          |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed(jcc#Array_Index_oof_bound) 

          Push( SignExtendByte( byte[ arrayPtr + arrayIndex ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BASTORE                                                                             $4B |
        ' |     Meaning: Store into signed byte array.                                                  |
        ' |     Description: An integer is popped off the stack, converted to a signed byte,            |
        ' |     and stored in the array at the index. The arrayref must be a reference to an            |
        ' |     array of signed bytes, and the index into the array must be an integer. If the          |
        ' |     integer value is too large to be a signed byte, it is truncated.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )
          
          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed(jcc#Array_Index_oof_bound) 

          if value <> SignExtendByte( value )
            Crashed(jcc#Signed_Int_too_big)
             
          byte[ arrayPtr + arrayIndex ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC                                                                                 $4C |
        ' |     Meaning: Push item from constant pool.                                                  |
        ' |     Description: byte1 is used as an unsigned 8-bit index into the constant pool            |
        ' |     of the current class. After the item at that index is resolved, it is pushed            |
        ' |     onto the stack.                                                                         |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC:

          LoadConstant( 1, Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC_W                                                                               $4D |
        ' |     Meaning: Push 2-byte item from constant pool.                                           |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the item at that index is            |
        ' |     resolved, it is pushed onto the stack.                                                  |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC_W:

          LoadConstant( 2, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC2_W                                                                              $4E |
        ' |     Meaning: Push long or double from constant pool.                                        |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the two-word constant at             |
        ' |     that index is resolved, it is pushed onto the stack.                                    |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC2_W:

          LoadConstant( 4, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IMUL, IDIV, IREM                                                                    $4F |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IMUL: Push( PopL *  PopR )
        jem#Jem_IDIV: Push( PopL /  PopR )
        jem#Jem_IREM: Push( PopL // PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     JSR                                                                                 $52 |
        ' |     Meaning: Jump subroutine.                                                               |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2. The                |
        ' |     address of the instruction immediately following the current instruction is             |
        ' |     pushed onto the stack to provide a return address from the subroutine.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_JSR:

          Push( pc )
          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     RET                                                                                 $53 |
        ' |     Meaning: Return from subroutine.                                                        |
        ' |     Description: Local variable in the current Java frame must contain a return             |
        ' |     address. The contents of the local variable are written into the program                |
        ' |     counter.                                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_RET:

          pc := GetVar( Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     TABLESWITCH                                                                         $54 |
        ' |     Meaning: Access jump table by index and jump.                                           |
        ' |     Description: Immediately after the tableswitch instruction, padding consisting          |
        ' |     of 0-3 zeros is inserted so that the next byte begins at an address that is a           |
        ' |     multiple of four. A series of offsets follows the padding. These offsets are            |
        ' |     signed 4-byte quantities and consist of a default offset, a low offset, and a           |
        ' |     high offset, followed by additional high - low + 1 offsets. These additional            |
        ' |     offsets are treated as a 0-based jump table.                                            |
        ' |     The index the offsets point to must be an integer. If the index is less than            |
        ' |     the low offset or greater than the high offset, the default offset is added to          |
        ' |     the address of the current instruction. Otherwise, the low offset is subtracted         |
        ' |     from the index, and the element at the position index - low offset in the jump          |
        ' |     table is extracted and added to the address of the current instruction.                 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_TABLESWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     LOOKUPSWITCH                                                                        $55 |
        ' |     Meaning: Access jump table by key match and jump.                                       |
        ' |     Description: Immediately after the lookupswitch instruction, padding                    |
        ' |     consisting of 0-3 zeros is                                                              |
        ' |     inserted so that the next byte begins at an address that is a multiple of               |
        ' |     four. A series of pairs of offsets follows the padding. These offsets are               |
        ' |     signed 4-byte quantities. The first item in the pair is the default offset, and         |
        ' |     the second item in the pair gives the number of pairs that follow. The                  |
        ' |     additional pairs consist of a match and an offset.                                      |
        ' |     The integer key on the stack is then compared against each of the matches. If           |
        ' |     the key is equal to one of the matches, the offset is added to the address of           |
        ' |     the current instruction. If the key does not match any of the matches, the              |
        ' |     default offset is added to the address of the current instruction.                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LOOKUPSWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |     Deal with unimplemened opcodes and forced PASM calls                                            |
' `-----------------------------------------------------------------------------------------------------'

PRI NotImplemented

  if opcode < 0
    if not allowNotImplementedNatives
      Crashed(jcc#Native_Method_not_implemented)
  else
    if mode == MODE_SPIN_ONLY
      if doJazzLib == 0
      else
        tv.Str( String( $0D,"PC " ) )
        tv.Hex( pc-bytecodeStart, 2 )
        tv.Str( String( " OP " ) )
        tv.Hex( byte[pc-bytecodeStart], 2 )
        Crashed(jcc#Spin_cant_handle_opcode)
    else
      if mode == MODE_SPIN_THEN_PASM
        if JvmPasm
          abort
        else
          Crashed(jcc#Spin_PASM_cant_handle_opcode)
      else
        if doJazzLib == 0
        else
          Crashed(jcc#PASM_Spin_cant_handle_opcode)

PRI UsePasm

  if JvmPasm
    abort
  else
    Crashed(jcc#Forced_PASM_cant_handle_opcode)

' .-----------------------------------------------------------------------------------------------------.
' |     Bytecode Operand Handling                                                                       |
' `-----------------------------------------------------------------------------------------------------'

PRI Operand_UnSignedByte : operand

  operand := byte[ pc++ ]
{
  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand, 2 )
}
  
PRI Operand_SignedByte : operand

  operand := SignExtendByte( Operand_UnSignedByte )

PRI Operand_UnSignedWord : operand

  operand := byte[ pc++ ] << 8
  operand |= byte[ pc++ ]
{
  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand,4)
}  
PRI Operand_SignedWord : operand

  operand := SignExtendWord( Operand_UnSignedWord )

PRI Operand_BytecodeAddress : operand

  operand := Operand_UnSignedWord + bytecodeStart

PRI SignExtendByte( n )

  result := ~ n

PRI SignExtendWord( n )

  result := ~~ n


' .-----------------------------------------------------------------------------------------------------.
' |     Stack Pushing                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PUB Push( n )
{ using stack from outside is risky. be careful. }

  if INT32
    PushLong( n )
  else
    PushWord( n )
    
PRI PushLong( n )

  stackAltered := true
  if INT32
    sp += 4
    if sp > @stack + ( STACK_SIZE << INTWORDS )
      Crashed(jcc#Stack_overflow)
    long[ sp ] := n
  else
    Crashed( String( "Cannot use PushLong if not INT32" ) )
   
PRI PushWord( n )

  stackAltered := true
  sp += 2
  if sp > @stack + ( STACK_SIZE << INTWORDS )
    Crashed(jcc#Stack_overflow)
  word[ sp ] := n

PRI PushAddr( n )

  if INT32
    PushLong( n )
  else
    if ChipVer == 1
      PushWord( n )
    else
      PushWord( n >> 16 )
      PushWord( n       )
      
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Popping                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PUB Pop
{ using stack from outside is risky. be careful. }

  if INT32
    result := PopLong
  else
    result := PopWord

PRI PopLong

  stackAltered := true
  if INT32
    if sp < @stack
      Crashed(jcc#Stack_underflow)
    result := long[ sp ]
    sp -= 4
  else
    Crashed(jcc#Cant_use_PopLong_if_not_INT32)
    
PRI PopWord

  stackAltered := true
  if sp < @stack
    Crashed(jcc#Stack_underflow)
  result := SignExtendWord( word[ sp ] )
  sp -= 2

PRI PopAddr

  if INT32
    result := PopLong
  else
    if ChipVer == 1
      result := PopWord
    else
      result := PopWord
      result |= PopWord << 16

PRI Tos

  result := Pop
  Push( result )
        
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Manipulation                                                                              |
' `-----------------------------------------------------------------------------------------------------'

PRI PopL : lhs | rhs

  '       .---.           .---.
  '       | L |     SP -->| R |
  '       |---|           `---'
  ' SP -->| R |
  '       `---'

  rhs := Pop
  lhs := Pop
  Push( rhs )

PRI PopR : rhs

  rhs := Pop

PRI Pop2

  Pop
  Pop

PRI Dup | a

  '       .---.           .---.
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                 SP -->| A |
  '                       `---'

  a := Pop

  Push( a )
  Push( a )

PRI Dup_X1 | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |           |---|
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop

  Push( a )
  Push( b )
  Push( a )

PRI Dup_X2 | a, b, c

  '       .---.           .---.
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | C |
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop
  c := Pop

  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2 | a, b

  '       .---.           .---.
  '       | B |           | B |
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop

  Push( b )
  Push( a )
  Push( b )
  Push( a )

PRI Dup2_X1 | a, b, c

'       .---.           .---.
'       | C |           | B |
'       | B |           | A |
' SP -->| A |           |---|
'       `---'           | C |
'                       | B |
'                 SP -->| A |
'                       `---'

  a := Pop
  b := Pop
  c := Pop

  Push( b )
  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2_X2 | a, b, c, d

  '       .---.           .---.
  '       | D |           | B |
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | D |
  '       `---'           | C |
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop
  c := Pop
  d := Pop

  Push( b )
  Push( a )
  Push( d )
  Push( c )
  Push( b )
  Push( a )

PRI Swap | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |     SP -->| B |
  '       `---'           `---'

  a := Pop
  b := Pop

  Push( a )
  Push( b )

' .-----------------------------------------------------------------------------------------------------.
' |    Variable Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'

PUB GetVar( offset )

  result := Peek( bp, offset )

PRI PutVar( offset, n )

  Poke( bp, offset, n )

PRI IncVar( offset, n )

  PutVar( offset, GetVar( offset ) + n )
  
PRI Peek( base, offset )

  if INT32
    result := long[ base + ( offset << 2 ) ]
  else
    result := SignExtendWord( word[ base + ( offset << 1 ) ] )
{
  if 0'jazzDebug
    tv.Str( string($d,"Peek "))
    tv.Hex( base, 4 )
    tv.Out( " ")
    tv.Hex( offset, 4 )
    tv.Str( string(" rc "))
    tv.Hex( result, 4 )
}
PRI Poke( base, offset, n )

  if INT32
    long[ base + ( offset << 2 ) ] := n
  else
    word[ base + ( offset << 1 ) ] := n
{
  if 0'jazzDebug
    tv.Str( string($d,"Poke "))
    tv.Hex( base, 4 )
    tv.Out( " ")
    tv.Hex( offset, 4 )
    tv.Str( string(" val "))
    tv.Hex( n, 4 )
}
' .-----------------------------------------------------------------------------------------------------.
' |    Arrays                                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first data item
' The Type and Bytlen are always Int(16) or Int(32) depending upon ChipVer *NOT* INT32 
' The array data entity size depends upon Typ

PRI NewArrayRef( typ, len )

  result := GetFromHeap( typ, len )
    
PRI PopArrayRef( typ, atPtr, atLen ) | arrayTyp, bytLen, datPtr

  result := Pop
{
  if jazzDebug
    tv.Str(string($d,"PopArrayRef "))
    tv.Str(jem.TypeNamePtr(typ))
    tv.Str(string(" refptr "))
    tv.Hex(result, 4)
    tv.Out(" ")
}
  if result == 0
    Crashed(jcc#Got_Null_Pointer_for_Array_Ref)
  else
     
    GetHeapInfo( result, @arrayTyp, @bytLen, @datPtr )
{
    if jazzDebug
      tv.Str(string(" type "))
      tv.Hex(arrayTyp, 4)
}
    long[ atPtr ] := datPtr
    long[ atLen ] := bytLen >> BytesShiftPerType( arrayTyp )
{
    if jazzDebug
      tv.Str(string(" array "))
      tv.Hex(word[atPtr], 4)
      tv.Str(string(" len "))
      tv.Hex(word[atLen], 4)
}
    if arrayTyp <> typ
      if typ <> jem#T_ANY
        Crashed(jcc#Incorrect_Array_Type)
     
' .-----------------------------------------------------------------------------------------------------.
' |    Objects                                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PRI PopObjectRef

  NotImplemented
  
' .-----------------------------------------------------------------------------------------------------.
' |    Heap Management                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first byte of data
' The Type and Bytlen are always Int(16) or Int(32) depending upon INT32
' The array data entity size depends upon Typ

PRI GetFromHeap( typ, len ) | bytLen, i

  bytLen := len << BytesShiftPerType(typ)
{  
  if jazzDebug
    tv.Out($0D)
    tv.Str( String( "GetFromHeap Typ  " ) )
    tv.Hex( typ,4 )
    tv.Out( " ")
    tv.Str( jem.TypeNamePtr(typ) )
    tv.Str( String( " Len  " ) )
    tv.Hex( len, 4 )
    tv.Str( String( " BytLen " ) )
    tv.Hex( bytLen, 4 )
}
  ' Pause_Ms( 10_000 )

  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    heap &= $FFFF_FFFE
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_WORD )
    word[ heap + ( HEAP_HEAD_TYPE << HEAP_WORD ) ] := typ
    word[ heap + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    heap &= $FFFF_FFFC
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_LONG )
    long[ heap + ( HEAP_HEAD_TYPE << HEAP_LONG ) ] := typ
    long[ heap + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_LONG )

  repeat bytLen
    byte[ i++ ] := 0
{
  if jazzDebug
    tv.Str(string(" heap "))
    tv.Hex(heap, 4)
}
  return heap

PRI GetHeapInfo( ptr, atTyp, atLen, atDat ) : nextPtr | typ, Bytlen, datPtr, ii, ts
 
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    typ    := word[ ptr + ( HEAP_HEAD_TYPE << HEAP_WORD ) ]
    bytLen := word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_WORD )
    nextPtr := datPtr + bytLen
    if nextPtr & 1
      nextPtr++
  else
    typ    := long[ ptr + ( HEAP_HEAD_TYPE << HEAP_LONG ) ]
    bytLen := long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_LONG )
    nextPtr := datPtr + bytLen
    if nextPtr & %11
      nextPtr := ( nextPtr | %11 ) + 1

  if atTyp
    long[ atTyp ] := typ
  if atLen
    long[ atLen ] := bytLen
  if atdat
    long[ atDat ] := datPtr

{
  ts := jem.TypeNamePtr(typ) 
  if byte[ts] <> "U"
    nextPtr := HEAP_BASE 'let ShowHeap know to stop
    
  if 0'jazzDebug
    tv.Out($0D)
    tv.Str(String( "GetHeapInfo  " ))
    tv.Hex(ptr,4)
    tv.Out(" ")
    ts := jem.TypeNamePtr(typ) 
    tv.Str(ts)
    if byte[ts] <> "U"
      tv.Str(String(" Len  "))
      if typ == jem#T_BYTE
        tv.Dec(bytLen)
      else
        tv.Dec(bytLen/2)
      tv.Str(String(" Dat "))
      tv.hex(datPtr,4)
      repeat ii from 0 to bytLen
        if typ == jem#T_BYTE
          if (ii // 16) == 0
            tv.out($d)
          tv.hex(byte[datPtr+ii],2)
        else         
          if (ii // 32) == 0
            tv.out($d)
          tv.hex(word[datPtr+ii++],4)
        tv.out(" ")
    else
      ShowHeap
      Crashed(jcc#GetHeapInfo_lookup_failed)
}      
      
PRI BytesShiftPerType( typ )

  case typ & jem#T_ANY
  
    jem#T_BOOLEAN: result := ( >| 1        ) - 1
    jem#T_CHAR:    result := ( >| CHRBYTES ) - 1
    jem#T_FLOAT:   result := ( >| 2        ) - 1
    jem#T_DOUBLE:  result := ( >| 2        ) - 1
    jem#T_BYTE:    result := ( >| 1        ) - 1
    jem#T_SHORT:   result := ( >| 2        ) - 1
    jem#T_INT:     result := ( >| INTBYTES ) - 1
    jem#T_LONG:    result := ( >| LNGBYTES ) - 1
    jem#T_ANY:     result := ( >| INTBYTES ) - 1
    
    other:
      tv.out($D)
      tv.hex(typ,4)
      Crashed(jcc#BytesShiftPerType_Unknown)

PRI BytesPerType( typ )

  case typ & jem#T_ANY
  
    jem#T_BOOLEAN: result := 1
    jem#T_CHAR:    result := 2
    jem#T_FLOAT:   result := 4
    jem#T_DOUBLE:  result := 4
    jem#T_BYTE:    result := 1
    jem#T_SHORT:   result := 2
    jem#T_INT:     result := 2
    jem#T_LONG:    result := 4
    jem#T_ANY:     result := 4
    
    other:
      tv.out($D)
      tv.hex(typ,4)
      Crashed(jcc#BytesPerType_Unknown_Type)

PRI TypePerSize( size )

  case size
    1: result := jem#T_BYTE
    2: result := jem#T_SHORT
    4: result := jem#T_INT 
    8: result := jem#T_LONG
    other: result := 2

PRI isHeapPtr(ptr) | p
  p := heap
  repeat while (p < HEAP_BASE)
    p := GetHeapInfo(p, 0, 0, 0)
    if p == ptr
      return 1
  result := 0
    
PUB GetHeapValuePtr(ptr, index) | type, len, data
{{
/**
 * get pointer from heap pointer[index]
 * @param ptr - the heap pointer
 * @param index - the heap address offset data index
 * @returns pointer of ptr[index]
 */
GetHeapValuePtr(ptr, index);
}}
  if isHeapPtr(ptr)
    GetHeapInfo(ptr, @type, @len, @data)
    return GetHeapDataPtr(data, type, len, index)
  result := 0  

PUB GetHeapValue(ptr, index) | type, len, data, shift
{{
/**
 * get value from heap pointer[index]
 * @param ptr - the heap pointer
 * @param index - the heap address offset data index
 * @returns value at ptr[index]
 */
GetHeapValue(ptr, index);
}}
  if isHeapPtr(ptr)
    GetHeapInfo(ptr, @type, @len, @data)
   if (index => 0) and (index < len<<BytesShiftPerType(type))
     return GetHeapDataValue(data, type, len, index)
   else
     tv.out($d)
     tv.dec(index)
     Crashed(jcc#GetHeapValue_index_oof_range)
  result := 0

PRI SetHeapValue(ptr, index, val) | type, len, data, shift
{{
/**
 * set value at heap pointer[index]
 * @param ptr - the heap pointer
 * @param index - the heap address offset data index
 * @param value - value to set at ptr[index]
 */
SetHeapValue(ptr, index, val);
}}
  GetHeapInfo(ptr, @type, @len, @data)
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    SetHeapDataValue(data, type, len, index, val)
  else
    tv.out($d)
    tv.dec(index)
    tv.out(">")
    tv.dec(len)
    Crashed(jcc#GetHeapValue_index_oof_range)

PRI GetHeapDataPtr(dataptr, type, len, index) | shift
{{
/**
 * get value from heap data[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @returns ptr of ptr[index]
 */
GetHeapDataPtr(dataptr, type, len, index);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: result := dataptr+index<<shift
      jem#T_CHAR:    result := dataptr+index<<shift
      jem#T_FLOAT:   result := dataptr+index<<shift
      jem#T_DOUBLE:  result := dataptr+index<<shift
      jem#T_BYTE:    result := dataptr+index<<shift
      jem#T_SHORT:   result := dataptr+index<<shift
      jem#T_INT:     result := dataptr+index<<shift
      jem#T_LONG:    result := dataptr+index<<shift
      jem#T_ANY:     result := dataptr+index<<shift
      other:         result := 0
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("GetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(result,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(jcc#GetHeapDataVal_index_oof_range)

PRI GetHeapDataValue(dataptr, type, len, index) | shift
{{
/**
 * get value from heap data[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @returns value at ptr[index]
 */
GetHeapDataValue(dataptr, type, len, index);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: result := word[dataptr+index<<shift]
      jem#T_CHAR:    result := word[dataptr+index<<shift]
      jem#T_FLOAT:   result := word[dataptr+index<<shift]
      jem#T_DOUBLE:  result := long[dataptr+index<<shift]
      jem#T_BYTE:    result := byte[dataptr+index<<shift]
      jem#T_SHORT:   result := word[dataptr+index<<shift]
      jem#T_INT:     result := word[dataptr+index<<shift]
      jem#T_LONG:    result := long[dataptr+index<<shift]
      jem#T_ANY:     result := word[dataptr+index<<shift]
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("GetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(result,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(jcc#GetHeapDataVal_index_oof_range)

PRI SetHeapDataValue(dataptr, type, len, index, val) | shift
{{
/**
 * set value at heap dataptr[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @param value - the value to set at dataptr[index]
 */
SetHeapDataValue(dataptr, type, len, index, val);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: word[dataptr+index<<shift] := val
      jem#T_CHAR:    word[dataptr+index<<shift] := val
      jem#T_FLOAT:   word[dataptr+index<<shift] := val
      jem#T_DOUBLE:  long[dataptr+index<<shift] := val
      jem#T_BYTE:    byte[dataptr+index<<shift] := val
      jem#T_SHORT:   word[dataptr+index<<shift] := val
      jem#T_INT:     word[dataptr+index<<shift] := val
      jem#T_LONG:    long[dataptr+index<<shift] := val
      jem#T_ANY:     word[dataptr+index<<shift] := val
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("SetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(val,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(jcc#SetHeapDataVal_index_oof_range)
         
' .-----------------------------------------------------------------------------------------------------.
' |    Constant Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'
{
PRI LoadConstant( constantSizeInBytes, tableIndex ) | constAdr, constLen, constOff, heapPtr, heapData, ii

  '              ; String index table.
  ' 0B25: dw $0B ; String offset hi = $0B
  ' 0B26: dw $2B ;               lo = $2B
  ' 0B27: dw $0B ; String offset hi = $0B
  ' 0B28: dw $35 ;               lo = $35
  ' 0B29: dw $0B ; String offset hi = $0B
  ' 0B2A: dw $43 ;               lo = $43

  constAdr := byte[ ct + tableIndex << 1 + 0 ] << 8
  constAdr |= byte[ ct + tableIndex << 1 + 1 ] 
  constAdr += bytecodeStart 
  
  ' 0B35: dw $01 ; String class  hi = $01
  ' 0B36: dw $C9 ;               lo = $C9
  ' 0B37: dw $00 ; String length hi = $00
  ' 0B38: dw $04 ;               lo = $04
  ' 0B39: dw $0B ; Pointer       hi = $0B
  ' 0B3A: dw $3B ;               lo = $3B
  ' 0B3B: dw $01 ; Object class  hi = $01
  ' 0B3C: dw $20 ;               lo = $20
  ' 0B3D: dw $00 ; String length hi = $00
  ' 0B3E: dw $04 ;               lo = $04
  ' 0B3F: dw "n" ; String Data   "null"
  ' 0B40: dw "u"
  ' 0B41: dw "l"
  ' 0B42: dw "l"

  mode := 0 ' We messed with stuff so don't let PASM try !
  NotImplemented
}

' .-----------------------------------------------------------------------------------------------------.
' |    Flow Control                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI BranchIf ( condition ) | offset

  offset := Operand_SignedWord - 3
{
  if debugging
    tv.Out( " " )
    tv.Out( "(" )
    tv.Out( " " )
    tv.Dec( pc + offset - mb )
    tv.Out( " " )
    tv.Out( ")" )
}  
  if condition
    pc += offset

PRI InvokeProgram( bytecodeStartPtr, startMain ) |  constantTable, mainMethodAdr, debstate, numStaticFields, mainMethodBase
 
  bytecodeStart := bytecodeStartPtr
  
  ' 0000: dw $A4 ; Placeholder for run/debug.
  ' 0001: dw $0  ; Placeholder for the length.
  ' 0002: dw $01 ; Version: 0.1
  ' 0003: dw 16  ; Number of static fields = 8
  '
  ' 0004: dw $00 ; `main' method: Offset hi = $00
  ' 0005: dw $1E ;                       lo = $1E
  '
  ' 0006: dw $01 ; java.lang.Object class: Offset hi = $01
  ' 0007: dw $12 ;                                lo = $12
  '
  ' 0008: dw $01 ; java.lang.String class: Offset hi = $01
  ' 0009: dw $BB ;                                lo = $BB
  '
  ' 000A: dw $00 ; Throwable.throwVMException method: Offset hi = $00
  ' 000B: dw $36 ;                                           lo = $36
  '
  ' 000C: dw $0B ; String table offset hi = $0B
  ' 000D: dw $03 ;                     lo = $03
  '
  ' 000E: dw $00 ; Callback method offset hi = $00
  ' 000F: dw $00 ;                        lo = $00
  '
  ' 0010: dw $01 ; java.lang.OutOfMemoryError class: Offset hi = $01
  ' 0011: dw $95 ;                                          lo = $95
  '
  ' 0012: dw $0B ; Stack base hi = $0B
  ' 0013: dw $40 ;            lo = $40

{  
  if jazzDebug
    tv.Str(String($d,"bytecodeStart $"))
    tv.hex(bytecodeStart,4)
    tv.Str(String(" version "))
    tv.hex(byte[ bytecodeStart + $0002 ],2)
}
  case byte[ bytecodeStart + $0002 ] ' File Version

    ' Standard Javelin Int(16), no Long
    $01   : if INT32
              tv.Str( String( $0D,"Jem INT16's auto'd to INT32" ) )
            jemFileCountAdjust := 1
            
'== ' Int(16), Long(32)
'== $80   :  if INT32
'==           tv.Str( String( $0D,"Jem File INT16's auto-promoted to INT32" ) )
    
'== ' Int(32), Long(32)
'== $C0   : if not INT32
'==           Crashed(jcc#Jem_File_requires_INT32)
            
    other :
      tv.hex(byte[ bytecodeStart + $0002 ],2)
      Crashed(jcc#Bad_Jem_File_Version)

  'tv.Str(String($d,"bytecodeStart $"))
  'tv.hex(bytecodeStart,4)

  gBytecodeLen    := byte[ bytecodeStart + $0001 ] << 8

  numStaticFields := byte[ bytecodeStart + $0003 ] >> jemFileCountAdjust

  mainMethodAdr  := byte[ bytecodeStart + $0004 ] << 8
  mainMethodAdr  |= byte[ bytecodeStart + $0005 ]
  'gBytecodeMain  := mainMethodAdr 

  gObjectClassId := byte[ bytecodeStart + $0006 ] << 8
  gObjectClassId |= byte[ bytecodeStart + $0007 ] 
  
  constantTable  := byte[ bytecodeStart + $000C ] << 8
  constantTable  |= byte[ bytecodeStart + $000D ] 

  
  mainMethodBase := byte[ mainMethodAdr + bytecodeStart + 3] << 8
  mainMethodBase |= byte[ mainMethodAdr + bytecodeStart + 4]
  gBytecodeMain  := mainMethodBase 

{
  if 1'jazzDebug
    tv.Str(String($d,"bytecodeStart $"))
    tv.hex(bytecodeStart,4)
    
    tv.Str(String($d,"bytecodeLen $"))
    tv.hex(gBytecodeLen,4)

    tv.Str(String($d,"numStaticFields $"))
    tv.hex(numStaticFields,4)
   
    tv.Str(String($d,"mainMethodAdr $"))
    tv.hex(mainMethodAdr,4)
  
    tv.Str(String($d,"gObjectClassId $"))
    tv.hex(gObjectClassId,4)
  
    tv.Str(String($d,"constantTable $"))
    tv.hex(constantTable,4)

    tv.Str(String($d,"gBytecodeMain $"))
    tv.hex(gBytecodeMain,4)
}     
    
  mainMethodBase += bytecodeStart

  debstate := jazzDebug
  jazzDebug := 0  
  LoadClassInfo(mainMethodBase)
  
{ LoadClassInfo accessor function tests
  LookupClassIdIndex($118)
  LookupMethodIndex($118,$11B)
  LookupMethodClassId($11B)
}
  jazzDebug := debstate

  ' Create a chunk which we can hold our static fields in, and point to the fields ...

  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    vp := GetFromHeap( jem#T_INT, numStaticFields) + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    vp := GetFromHeap( jem#T_INT, numStaticFields) + ( HEAP_HEAD_DATA << HEAP_LONG )
  
  ct := constantTable + bytecodeStart

  InitConstantTable
  'gMainObject := GetNewClass($0014)

  gCallMainMethodAddr := mainMethodAdr + bytecodeStart

  if startMain 
    CallMethod( mainMethodAdr + bytecodeStart, false )

    
PRI InvokeStatic( methodBlockAdr )

  CallMethod( methodBlockAdr, false )

PRI CallMethod( methodBlockAdr, extra ) | classid, frameStackSize, numberOfParameters, numberOfLocals, flags, methodAdr, exceptionTable, i, oldFramePointer

  '              ; Method = 'MyMethod'
  ' 004E: dw 4   ; 0 - Number of parameters = 2
  ' 004F: dw 6   ; 1 - Number of locals = 3
  ' 0050: dw $00 ; 2 - Flags     = $00
  ' 0051: dw $04 ; 3 - Offset hi = $04
  ' 0052: dw $E8 ; 4 - Offset lo = $E8
  ' 0053: dw $00 ; 5 - Empty exception table hi = $00
  ' 0054: dw $00 ; 6 - Empty exception table lo = $00

  numberOfParameters := byte[ methodBlockAdr + 0 ] >> jemFileCountAdjust
  numberOfLocals     := byte[ methodBlockAdr + 1 ] >> jemFileCountAdjust
  flags              := byte[ methodBlockAdr + 2 ]
  methodAdr          := byte[ methodBlockAdr + 3 ] << 8
  methodAdr          |= byte[ methodBlockAdr + 4 ]
  exceptionTable     := byte[ methodBlockAdr + 5 ] << 8
  exceptionTable     |= byte[ methodBlockAdr + 6 ]


  if sp > @stack
    oldFramePointer := fp
    fp := sp
  else
    oldFramePointer := 0
    fp := 0
    
  if callDebug
    tv.Str( string($d,"CallMeth "))
    tv.Hex( methodAdr, 4 )
    tv.Str( string(" Blk@ "))
    tv.Hex( methodBlockAdr-bytecodeStart, 4 )
    tv.Str( string(" #P "))
    tv.Hex( numberOfParameters, 2 )
    tv.Str( string(" #L "))
    tv.Hex( numberOfLocals, 2 )
    tv.Str( string(" ET "))
    tv.Hex( exceptionTable, 4 )
    tv.Str( string(" Fl "))
    tv.Hex( flags, 2 )
    tv.Str( string(" oFP "))
    tv.Hex( oldFramePointer, 4 )

    
  '               SP -->
  '       .----.        .----.          .----.          .----.
  '       | A  |        |    |          | SP |          | SP |
  ' SP -->| B  |        :    :          :    :          :    :
  '       `----'        |    |          | BP |          | BP |
  '                     :    :          :    :          :    :
  '                     |    |          | PC |          | PC |
  '                     |----|          |----|          |----|
  '                     | A  |    BP -->| A  |    BP -->| A  |
  '                     | B  |    SP -->| B  |          | B  |
  '                     `----'          `----'          |----|
  '                                                     | 0  |
  '                                                     | 0  |
  '                                               SP -->| 0  |
  '                                                     `----'
  '
  ' 1) Adjust SP to below parameters, shuffle parameters up
  ' 2) Insert stack frame, adjust BP and SP
  ' 3) Push zeroed local variables

  sp -= ( (numberOfParameters) << INTWORDS )

  ' For InvokeVirtual, we need to keep a record of the working class.
  ' The working class excludes java.lang.Object and init.
  ' During InvokeVirtual, the method to invoke is represented as
  ' an offset from the class address or classid.
  '
   
  gMethodId := methodBlockAdr-bytecodeStart
  gClassId := word[fp+1]
  
  if callDebug
    classid := LookupMethodClassId(methodBlockAdr-bytecodeStart)
    tv.Str( string(" gCID $"))
    tv.Hex( classid, 4 )
    tv.Str( string(" gMId $"))
    tv.Hex( gMethodId, 4 )

      
  if numberOfParameters <> 0

    frameStackSize := ( ( @sp - @pc ) >> 2 ) + 1 ' How many PushAddr's follow

    if not INT32
      if ChipVer <> 1           
        frameStackSize <<= 1
                                
    repeat i from 1 to numberOfParameters           
      Poke( sp, i + frameStackSize, Peek( sp, i ) )
  
  repeat i from @sp to @pc step 4 ' Not -4 ! That's a Propeller Gotcha
    PushAddr( long[ i ] )
                  
  bp := sp + INTBYTES

  sp += ( numberOfParameters << INTWORDS )

  repeat numberOfLocals
    Push( 0 )

  ' peter's frame info
  '
  Push(gMethodId)              
  Push(oldFramePointer)
  Push(pc-bytecodeStart)

  et := exceptionTable + bytecodeStart

  case flags
  
    $00 :   pc := methodAdr + bytecodeStart
            mb := pc
            
            if 0'hasClinit
              repeat while byte[ mb ] == jem#Jem_INVOKESTATIC
                mb += 3
              pc := mb ' ****** HACK **** This skips all "invokestatic <??.<clinit>>" at start of 'main'

    $80:    gNativeCode := methodAdr
            ' let native code know the number of parameters being passed
            ' number of locals being used for native is always zero
            Push(numberOfParameters)
            if debugging
              ShowStack
{
    $80 :   if NativeMethod( methodAdr, numberOfParameters )
              ReturnFromMethodWithValue
            else
              ReturnFromMethod
}

    other :
      tv.out($d)
      tv.out($d)
      tv.hex(flags, 2)
      Crashed(jcc#Unknown_Method_Table_Flag)
  
PRI ReturnFromMethodWithValue

  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |                  | ?? |  <--.   |
  '       |----|                  |----|     |   |
  '       | SP |------.     SP -->| XX |     |   |
  '       :    :      |           `----'     |   |
  '       | BP |---.  |                      |   |
  '       :    :   |  `----------------------'   |
  '       | PC |   |                             |
  '       |----|   `-----------------------------'
  ' BP -->| A  |
  '       | B  |
  '       |----|
  ' SP -->| XX |
  '       `----'
  
  result := Pop
  ReturnFromMethod
  Push( result )
{
  if jazzDebug
    tv.Str( string($d," .. ReturnFromMethod with value "))
    tv.Hex(result,4)
}
      
PRI ReturnFromMethod | i, oldpc, oldfp, oldmd

  ' Note that it would not be essential to have pushed the SP in the current implementation as the SP
  ' is always the address preceding the stack frame and could therefore be determined as an offset from
  ' BP ( and also SP would point to where it should be when PC, MB and BP have been popped ) but this
  ' is designed to pre-empt stack frames having been allocated from the global heap at some time in the
  ' future.
  ' 
  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |            SP -->| ?? |  <--.   |
  '       |----|                  `----'     |   |
  '       | SP |-----------------------------'   |
  '       :    :                                 |
  '       | BP |---------------------------------'
  '       :    :
  '       | PC |
  '       |----|
  ' BP -->| A  |
  ' SP -->| B  |
  '       `----'

  ' pop peter's 
  oldpc := Pop 'from Push(pc)
  oldfp := Pop 'from Push(oldFramePointer)
  oldmd := Pop 'from Push(gMethodId)              
  
  sp := bp - INTBYTES

  repeat i from @pc to @sp step 4
    long[ i ] := PopAddr
{
  if jazzDebug
    tv.Str( string($d," .. ReturnFromMethod pc $"))
    tv.Hex(oldpc,4)
    tv.Str( string(" .. OldFP $"))
    tv.Hex(oldfp,4)
    tv.Str( string(" .. Method $"))
    tv.Hex(oldmd,4)
}
    
' .-----------------------------------------------------------------------------------------------------.
' |     Call Jvm PASM Handler                                                                           |
' `-----------------------------------------------------------------------------------------------------'

PRI JvmPasm

  result := false
  
' .-----------------------------------------------------------------------------------------------------.
' |     Utility Code                                                                                    |
' `-----------------------------------------------------------------------------------------------------'

PUB ShowStack | len, ii

  'if stackAltered ' dump it all
    stackAltered := false
    tv.Out( $0D )
    tv.str(string("ST "))
    tv.Hex( sp, 4 )
    tv.Out( " " )
    len := (( sp - @stack ) >> INTWORDS) + 1
    tv.Dec( len)
    'tv.Hex( ( ( sp - @stack ) >> INTWORDS ) + 1, 2)
    tv.out( " " )
    tv.out( $d )
    if len < 0
      len := 15
    if  INT32
      repeat ii from 0 to len/2 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackLong( ii )
    else
      repeat ii from 0 to len 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackWord( ii )
  
PRI showStackLong( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 2 ) => depth
    tv.Hex( long[ sp - (depth<<2) ], 8 )
  else
    tv.Str( String("--------") )

PRI showStackWord( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 1 ) => depth
    tv.Hex( word[ sp - (depth<<1) ], 4 )
  else
    tv.Str( String("----") )

PUB PrintStack(len) | ii
    if  INT32
      repeat ii from 0 to len/2 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackLong( ii )
    else
      repeat ii from 0 to len 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackWord( ii )
  
PUB ShowHeap | ptr, typ, bytLen, datPtr, nextPtr

  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, 0 )
    ShowHeapEntry( ptr, typ, bytLen )
    ptr := nextPtr
    
PRI ShowHeapEntry( ptr, typ, bytLen )

  tv.Out( $0D )
  tv.Hex( ptr, 4 )
  tv.Out( "-" )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_WORD ) - 1, 4 )
  else
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_LONG ) - 1, 4 )
  tv.Out( " " )
  tv.Dec( typ )
  if typ < 10
    tv.Out( " " )
  tv.Out( " " )
  tv.Str( jem.TypeNamePtr(typ) )
  tv.Dec( BitSizePerType(typ) ) 
  tv.Out( "[" )
  tv.dec( bytLen >> BytesShiftPerType(typ) )
  tv.Out( "]" )  
  tv.Out( " " )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Dec( HEAP_HEAD_DATA << HEAP_WORD )
  else
    tv.Dec( HEAP_HEAD_DATA << HEAP_LONG )
  tv.Out( "+" )
  tv.Dec( bytLen )
  tv.Str( String( " Bytes" ) )
  if typ & $80
    tv.Str( String( " DEAD" ) )
    
PRI BitSizePerType(typ)

  result := 8 << BytesShiftPerType(typ)

PRI CompactHeap | ptr, typ, bytLen, datPtr, nextPtr, nextTyp
{
  This should work but not tested and never called
  
  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, @datPtr )
    if typ & jem#T_DELETED
      if ptr == heap
        heap := nextPtr
      else
        if nextPtr < HEAP_BASE
          GetHeapInfo( nextPtr, @nextTyp, 0, 0 )
          if nextTyp & jem#T_DELETED
            ' Combine the two
            nextPtr := GetHeapInfo( nextPtr, 0, 0, 0 )
            if ( ChipVer == 1 ) or ( not BIGARRAYS )
              word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := nextPtr - datPtr 
            else
              long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := nextPtr - datPtr  
            nextPtr := ptr
    ptr := nextPtr
}

               
PRI Crashed( crashcode )

  dumpRegs
  ShowStack
  'ShowHeap
  tv.Out( $0D )
  result := jcc.getCrashString(crashcode)
  if result
    tv.Str( result )
  else
    tv.str(string(" JVM Crashed - Code: "))
    tv.dec(crashcode)
    
  abort jvmCrashed
  repeat

PRI Pause_Ms( ms )

  waitcnt( CLKFREQ / 1000 * ms + CNT )


VAR ' gClassIdArray
  long gClassIdArray  ' heap pointer to class id array with id's stored by index in order found
  long gMethodIdArray ' heap pointer to array of method id arrays 
  
DAT ' ClassLoad method descriptor
  methNumParms          byte 0
  methNumLocals         byte 0
  methFlags             byte 0
  methOffset            word 0
  methExcTable          word 0
  methClass             word 0
  methAddr              word 0

CON
  methConsLen = 7               ' length of method constant 
  methDescLen = 4*2+3

PRI LoadClassInfo(mainMethodBase) | listIndex, superClass, classAddr, classFields, numClass, numMethod, methArray, data, type, len
{{
/**
 * Load a class info from the jem code binary file class table. This is an embedded
 * implementation so we don't look for files; everything is in the binary.
 * This table is used for classid lookup. ANEWARRAY is one code that needs this.
 *
 * The method builds two arrays: classid array and methodid array
 * The classid array is a list of class id's arranged by order found.
 * The methodid array is a list of arrays where each array is bound
 * to a class by classid index and contains method id's.
 *
 * @param mainMethodBase - this is pc address of the main method
 * @returns pointer to classlist heap "array"
 */
LoadClassInfo(mainMethodBase);
}}
  ' all classes have a super class except java.lang.Object and it's address is -1
  ' all classes have a fields field. all classes have at least one method.
{  
  if 0'jazzDebug
    tv.str(string($d,"LoadClassInfo $"))
    tv.hex(bytecodeStart+20,4)
    tv.str(string(" main addr $"))
    tv.hex(mainMethodBase,4)
}

  ' count classes
  '
  numClass := 0

  ' if we run into a new class, methExcTable will be non-zero.
  ' if we exceed class definitions, classAddr will be > mainMethodBase
  '
  ' in this loop we create stack entries. the end result will be
  ' classAddr0, methArray0, classAddr1, methArray1 ... classAddrN, methArrayN
  ' in the middle, methAddrs are pushed then popped to the methArrays
  ' 
  methExcTable := 0
  listIndex := bytecodeStart+20  ' class info starts at 20
  repeat while (listIndex < mainMethodBase)
    ' class address
    classAddr := listIndex-bytecodeStart
    Push(classAddr)
    superClass  := byte[listIndex++] << 8
    superClass  |= byte[listIndex++]
    classFields := byte[listIndex++]
{
    if 0'jazzDebug
      tv.str(string($d,"cAddr $"))
      tv.hex(classAddr,4)
      tv.str(string(" supC $"))
      tv.hex(superClass,4)
      tv.str(string(" cFlds "))
      tv.dec(classFields)
}     
    ' cound methods
    '
    numMethod := 0
    repeat while (listIndex < mainMethodBase) ' class info starts at 20
      bytemove(@methNumParms, listIndex, methConsLen)
      if methExcTable <> 0
        quit
      else
        methClass := classAddr
        methAddr := listIndex-bytecodeStart
        Push(methAddr)
        numMethod++
{
        if 0'jazzDebug
          tv.str(string($d,"// classSig $"))
          tv.hex(methClass,4)
          tv.str(string(" methAddr $"))
          tv.hex(methAddr,4)
          tv.str(string($d,"@methNumParms $"))
          tv.hex(@methNumParms,4)
          tv.str(string(" val $"))
          tv.hex(methNumParms,2)
          tv.str(string($d,"@methNumLocals $"))
          tv.hex(@methNumLocals, 4)
          tv.str(string(" val $"))
          tv.hex(methNumLocals, 2)
          tv.str(string($d,"@methFlags $"))
          tv.hex(@methFlags, 4)
          tv.str(string(" val $"))
          tv.hex(methFlags, 2)
          tv.str(string($d,"@methOffset $"))
          tv.hex(@methOffset,4)
          tv.str(string(" val $"))
          tv.hex(methOffset,4)
          tv.str(string($d,"@methExcTable $"))
          tv.hex(@methExcTable,4)
          tv.str(string(" val $"))
          tv.hex(methExcTable,4)
}
      listIndex += methConsLen

    methArray := GetFromHeap(jem#T_SHORT,numMethod) 
    GetHeapInfo(methArray, @type, @len, @data)

    repeat while(numMethod-- > 0)
      methAddr := Pop
      SetHeapDataValue(data, type, len, numMethod, methAddr)
{
      if 0'jazzDebug
          tv.str(string($d,"classNum "))
          tv.dec(numClass)
          tv.str(string(" classAddr $"))
          tv.hex(classAddr,4)
          tv.str(string(" methAddr $"))
          tv.hex(methAddr,4)
}
    Push(methArray)
{          
    if 0'jazzDebug
      tv.str(string($d,"cCount $"))
      tv.hex(numClass,4)
      tv.str(string(" mCount $"))
      tv.hex(numMethod,4)
}
    numClass++


  ' create storage for method ids.
  '
  gMethodIdArray := GetFromHeap(jem#T_SHORT,numClass)
{
  if jazzDebug 
    GetHeapInfo(gMethodIdArray, @type, @len, @data) ' debug only
}
  ' create storage for class ids
  '  
  result := GetFromHeap(jem#T_SHORT, numClass)
  GetHeapInfo(result, @type, @len, @data)

  ' put method arrays in class num indexed method array storage
  ' put classid = classAddr values in class list
  '  
  repeat while(numClass-- > 0)
    methArray := Pop
    SetHeapValue(gMethodIdArray, numClass, methArray)
    classAddr := Pop
    SetHeapDataValue(data, type, len, numClass, classAddr)
{
    if 0'jazzDebug
      tv.str(string($d,"classNum "))
      tv.dec(numClass)
      tv.str(string(" classAddr $"))
      tv.hex(classAddr,4)
}  
  GetHeapInfo(result, 0, 0, @data)
  gClassIdArray := result

  
PRI LookupClassIdIndex(classid) | ii, type, len, data, cid
{{
/**
 *
 * Get classid array index.
 * @param classid - classid is hex 4 signature at class bytecode address
 * @returns classid's 0..classnum or -1 on failure
 */
LookupClassIdIndex(classid);
}}
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupClassIdIndex $"))
    tv.hex(classid,4)
}
  result := -1 

  GetHeapInfo(gClassIdArray, @type, @len, @data)
  repeat ii from 0 to (len-1)>>BytesShiftPerType(type)
    cid := GetHeapDataValue(data, type, len, ii)
    'cid := GetHeapValue(gClassIdArray, ii)  ' another way
    'cid := word[data+ii<<1]                 ' easiest way to get it wrong
{
    if 0'jazzDebug
      tv.str(string($d,"Lookup $"))
      tv.hex(classid,4)
      tv.str(string(" classId $"))
      tv.hex(cid,4)
}
    if cid == classid
{
      if 0'jazzDebug
        tv.str(string($d,"/* LookupClassIdIndex $"))
        tv.hex(classid,4)
        tv.str(string(" = "))
        tv.dec(ii)
        tv.str(string(" */ "))
}
      return ii

PRI LookupClassAtIndex(index) | ii, type, len, data, cid
{{
/**
 *
 * Get classid array index.
 * @param index - class index of classid
 * @returns classid at index or 0 on failure
 */
LookupClassAtIndex(index);
}}
{
  if 1'jazzDebug
    tv.str(string($d,"// LookupClassAtIndex "))
    tv.dec(index)
}
  result := 0 

  GetHeapInfo(gClassIdArray, @type, @len, @data)
  if index < len and index => 0
    result := GetHeapDataValue(data, type, len, index)

PRI LookupMethodIndex(classid, methid) | ii, type, len, data, methArray, cid
{{
/**
 *
 * Get method array index in a class
 *
 * @param classid - classid is hex 4 signature at class bytecode address
 * @param methid  - methid  is hex 4 signature at method bytecode address
 *
 * @returns classid's method num 0..methonum or -1 on failure
 */
LookupMethodIndex(classid, methid);
}}
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupMethodIndex $"))
    tv.hex(classid,4)
    tv.str(string(" $"))
    tv.hex(methid,4)
}
  result := -1
   
  cid := LookupClassIdIndex(classid)
  if cid < 0
    return result
    
  GetHeapInfo(gMethodIdArray, @type, @len, @data)
  methArray := GetHeapDataValue(data, type, len, cid)
  if methArray == 0
    return result
  GetHeapInfo(methArray, @type, @len, @data)
  repeat ii from 0 to (len-1)>>BytesShiftPerType(type)
    result := GetHeapDataValue(data, type, len, ii)
{
    if 0'jazzDebug
      tv.str(string($d,"Lookup $"))
      tv.hex(classid,4)
      tv.str(string(" classId $"))
      tv.hex(result,4)
}      
    if result == methid
{
      if 0'jazzDebug
        tv.str(string($d,"/* LookupMethodIndex $"))
        tv.hex(classid,4)
        tv.str(string(" $"))
        tv.hex(methid,4)
        tv.str(string(" = $"))
        tv.hex(ii,4)
        tv.str(string(" */ "))
}
      return ii


PRI LookupMethodClassId(methid) | ii, jj, ctype, clen, cdata, mtype, mlen, mdata, methArray, type, len, data, cid
{{
/**
 * Get method's class id ... class id is bytecode offset of class definition.
 * @param methid  - methid  is hex 4 signature at method bytecode address
 * @returns classid's method num 0..methonum or -1 on failure
 */
LookupMethodClassId(methid);
}}
  result := 0
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupMethodClassId $"))
    tv.hex(methid,4)
}   
  GetHeapInfo(gClassIdArray, @ctype, @clen, @cdata)
  GetHeapInfo(gMethodIdArray, @mtype, @mlen, @mdata)
  repeat ii from 0 to (clen-1)>>BytesShiftPerType(type)
    methArray := GetHeapDataValue(mdata, mtype, mlen, ii)
    GetHeapInfo(methArray, @type, @len, @data)
    repeat jj from 0 to (len-1)>>BytesShiftPerType(type)
      result := GetHeapDataValue(data, type, len, jj)
{
      if 0'jazzDebug
        tv.str(string($d,"Lookup $"))
        tv.hex(ii,4)
        tv.str(string(" methId $"))
        tv.hex(result,4)
}
      if result == methid
        result := GetHeapDataValue(cdata, ctype, clen, ii)
{
        if 0'jazzDebug
          tv.str(string($d,"LookupMethodClassId result $"))
          tv.hex(result,4)
}
        return result

  
CON

  #0, ObjectThisClass, ObjectSuperClass, ObjectFields, ObjectLastEnum ' field enum should be last

{
Objects have this pointer, superclass pointer, constant table pointer (array of arrays), and field pointers.
}
  
PRI doNEW | object
{{
/**
 * See case jem#Jem_NEW comment for basic operation description.
 * Creates a new class object and puts reference on stack.
 */
doNEW();
}}     
  'classAddr := Operand_BytecodeAddress
  object := GetNewClass(Operand_BytecodeAddress)
  Push(object)
  
  result := 0
  
PRI GetNewClass(classAddr) | object, ii, fieldPtr, data, superClass, numFields, type, len
{{
/**
 * 
 * Create a new class object with classAddr resulting in a reference which is a heap pointer
 * with entries populated according to the CON #0 enum above ... ObjectThisClass, .....
 * Constant table is not loaded in each object ... this could change if necessary ...
 * constant table is a global array of String objects for LDC to access.
 *
 * @returns object reference
 */
GetNewClass(classAddr);
}}     

{
  if jazzDebug
    tv.Str( String($d,"New " ) )
    tv.Hex( pc-bytecodeStart, 4 )
}  
  superClass := byte[classAddr+0] << 8
  superClass |= byte[classAddr+1]
  numFields  := byte[classAddr+2] >> 1
   
  ' allocate heap pointer for class
  ' get enough storage for superClass and numFields
  ' fields are stored in the stack ???
  '
  object := GetFromHeap(jem#T_ANY, ObjectLastEnum) 
  GetHeapInfo(object, @type, @len, @data)
{ 
  if jazzDebug
    tv.Str(string($d,"NewClass pc "))
    tv.Hex(pc-bytecodeStart, 4 )
    tv.Str(string(" this "))
    tv.Hex(object, 4 )
    tv.Str(string(" class addr "))
    tv.Hex(classAddr, 4 )
    tv.Str(string(" SC "))
    tv.Hex(superClass, 4 )
    tv.Str(string(" F$"))
    tv.Hex(numFields, 2 )
    tv.Str(string(" t$"))
    tv.Hex(type, 4 )
    tv.Str(string(" l$"))
    tv.Hex(len, 2 )
} 
  'save this class name
  classAddr -= bytecodeStart
  SetHeapDataValue(data, type, len, ObjectThisClass, classAddr)
  SetHeapDataValue(data, type, len, ObjectSuperClass, superClass)
  fieldPtr := GetFromHeap(jem#T_ANY,numFields+1)
  SetHeapDataValue(data, type, len, ObjectFields, fieldPtr)
  repeat ii from 0 to numFields
    SetHeapValue(fieldPtr, ii, GetFromHeap(jem#T_ANY, 1))

  ' handle constant table globally if possible
  'SetHeapDataValue(data, type, len, ObjectConTable, GetConstantTable(object)) 'GetFromHeap(jem#T_ANY,GetConstantTableSize+1))
    
  result := object
  if jazzDebug
    ShowObject(result)

PRI GetConstantTableSize | constAddr
  constAddr := byte[ct + 0] << 8
  constAddr |= byte[ct + 1]
  constAddr += bytecodeStart 
  result := (constAddr - ct) / 2

PRI InitConstantTable | ii, jj, type, len, data, constAddr, constLen, constOff, constCnt, constSize, classAddr, ptr, object
  'if gConstTable <> 0
  '  return gConstTable
    
  constSize := 10
  constAddr := byte[ct + 0] << 8
  constAddr |= byte[ct + 1]
  constAddr += bytecodeStart 

  constCnt := (constAddr - ct) / 2
{
  if jazzDebug
    tv.Str(string($d,"InitCT "))
    tv.Dec(constCnt)
    tv.Str(string(" cadr "))
    tv.Dec(constAddr-bytecodeStart)
    tv.Out(" ")
    tv.Hex(constAddr-bytecodeStart, 4)
    tv.Str(string(" ct "))
    tv.Dec(ct-bytecodeStart)
    tv.Out(" ")
    tv.Hex(ct-bytecodeStart, 4)
}
  gConstTable := GetFromHeap(jem#T_ANY,constCnt)
  if gConstTable == 0
    Crashed(jcc#Cant_find_object_table_pointer)

  repeat ii from 0 to constCnt-1
    constLen := byte[constAddr + 2] << 8
    constLen |= byte[constAddr + 3]
    constOff := constAddr + constSize

    classAddr := byte[constAddr+0] << 8
    classAddr |= byte[constAddr+1]
{
    if 1'jazzDebug
      tv.Str(string($d,"InitCT "))
      tv.Dec(constCnt)
      tv.Str(string(" index "))
      tv.Dec(ii)
      tv.Str(string(" classadr "))
      tv.Hex(classAddr, 4 )
      tv.Str(string(" cadr "))
      tv.Hex(constAddr-bytecodeStart, 4 )
      tv.Str(string(" clen "))
      tv.Hex(constLen, 4 )
      tv.Str(string(" coff "))
      tv.Hex(constOff-bytecodeStart, 4 )
      tv.Out(" ")
      tv.Hex(constOff, 4 )
      tv.Out(" ")  
}
    object := GetNewClass(classAddr+bytecodeStart)

    ptr := GetFromHeap(jem#T_BYTE, constLen+1)
    
    GetHeapInfo(ptr, @type, @len, @data)
    SetObjectField(object, 0, classAddr)
    SetObjectField(object, 1, len)
    SetObjectField(object, 2, ptr)
{
    if 1'jazzDebug
      tv.Str(string(" len "))
      tv.Dec(len)
      tv.out(" ")
      'tv.tx(byte[constOff+jj])
}      
    repeat jj from 0 to constLen
      SetHeapDataValue(data, type, len, jj, byte[constOff+jj])

    ' put const in object's constant table ... this may be moved later
    SetHeapValue(gConstTable, ii, object)
    
    'ShowObject(object)
    constAddr += constSize + constLen

  result := gConstTable

PUB ShowObject(object) | ii, ptr, cptr, fptr, type, len, data, clen, flen, cdata, fdata
{{
/**
 * This displays all of objects contents if jazzDebug is enabled.
 * ObjectThisClass, ObjectSuperClass, ObjectFields ... from enum are displayed
 * @param object - object pointer
 */
__ void ShowObject(void* object);
}}
  'if jazzDebug
  'jazzDebug := 0
  GetHeapInfo(object, @type, @len, @data)  
  tv.Str(string($d," ShowObject type $"))
  tv.Hex(type, 4 )
  ptr := GetHeapDataValue(data, type, len, ObjectThisClass)
  tv.Str(string(" this $"))
  tv.Hex(ptr, 4)
  ptr := GetHeapDataValue(data, type, len, ObjectSuperClass)
  tv.Str(string(" Super $"))
  tv.Hex(ptr, 4 )
{    
  cptr := gConstTable 'GetHeapDataValue(data, type, len, ObjectConTable)
  tv.Str(string(" ConTable $"))
  tv.Hex(cptr, 4 )
  GetHeapInfo(cptr, 0, @clen, @cdata)
  repeat ii from 0 to clen>>1-1
    tv.out(" ")
    tv.hex(GetHeapDataValue(cdata, type, clen, ii),4)
}
  fptr := GetHeapDataValue(data, type, len, ObjectFields)
  tv.Str(string(" Fields $"))
  tv.Hex(fptr, 4 )
  GetHeapInfo(fptr, 0, @flen, @fdata)
  repeat ii from 0 to flen>>1-1
    tv.out(" ")
    tv.hex(GetHeapDataValue(fdata, type, len, ii),4)
  'jazzDebug := 1
 
              
PRI GetObjectField(object, index) | ptr, type, len, data 
{{
/**
 * gets object var field[index] value
 * @param object - object pointer
 * @param index - index to objects field
 * @returns value to from field[index]
 */
GetObjectField(objectref, index);
}}
  if jazzDebug
    ShowObject(object)
  ptr := GetHeapValue(object,ObjectFields)
  if ptr == 0
    Crashed(jcc#Cant_find_object_field_pointer)

  GetHeapInfo(ptr, @type, @len, @data)
  if index =< len  
    result := GetHeapDataValue(data, type, len, index)
  else
    tv.Str(string($d,"GetObjectField "))
    tv.Dec(index)
    tv.Str(string(" < len ? "))
    tv.Dec(len)
    Crashed(jcc#GetObjectField_index_oof_bound)
{
  if jazzDebug
    tv.Str(string($d,"GetObjectField "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(result, 4 )
}
           
PRI SetObjectField(object, index, value) | ptr, type, len, data
{{
/**
 * sets object var field[index] = value
 * @param object - object pointer
 * @param index - index to objects field
 * @param value - value to store in field[index]
 */
SetObjectField(objectref, index, value);
}} 
  if jazzDebug
    ShowObject(object)
{
  if jazzDebug
    tv.Str(string($d,"SetObjectField "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(value, 4 )
}
  ptr := GetHeapValue(object,ObjectFields)
  if ptr == 0
    tv.str(string("SetObjectField "))
    Crashed(jcc#Cant_find_object_field_pointer)
  GetHeapInfo(ptr, @type, @len, @data)
  if index =< len  
    SetHeapDataValue(data, type, len, index, value)
  else
    tv.Str(string($d,"SetObjectField "))
    tv.Dec(index)
    tv.Str(string(" < len ? "))
    tv.Dec(len)
    Crashed(jcc#SetObjectField_index_oof_bound)
             
PRI LoadConstant(constantSizeInBytes, tableIndex) | ii, constAdr, constLen, constOff, classAddr, objectref, ptr, type, len, data
{{
/**
 * This loads a constant ....
 * @todo evaluate how int, long, and double will be treated etc....
 * @param constantSizeInBytes -  using one of two methods ....
 *  If constant size == 1, load the String object from global constant table at tableIndex.
 *  If constant size > 1, load the constant offset value. LIKELY BROKEN.
 * @param tableIndex -  index of constant to load.
 * @returns 0 to let case know the function is implemented
 */
LoadConstant(constantSizeInBytes, tableIndex);
}}

  if gConstTable == 0
    Crashed(jcc#Cant_find_object_table_pointer)

  if constantSizeInBytes > 1
    constAdr := byte[ct + tableIndex << 1 + 0]
  else
    constAdr := byte[ct + tableIndex << 1 + 0] << 8
    constAdr |= byte[ct + tableIndex << 1 + 1]
     
  constAdr += bytecodeStart 

  constLen := byte[constAdr + 8] << 8
  constLen |= byte[constAdr + 9]
  constOff := constAdr + 10
{
  if jazzDebug
    tv.Str(string($d,"LDC "))
    tv.Dec(constantSizeInBytes)
    tv.Str(string(" "))
    tv.Hex(tableIndex, 4 )
    tv.Str(string(" cadr "))
    tv.Hex(constAdr-bytecodeStart, 4 )
    tv.Str(string(" clen "))
    tv.Hex(constLen, 4 )
    tv.Str(string(" coff "))
    tv.Hex(constOff-bytecodeStart, 4 )
    tv.Out(" ")
    tv.Hex(constOff, 4 )
    tv.Out(" ")  
}
  ' ldc2_w can only be a long/double
  '
  if constantSizeInBytes == 4
    Push(word[constOff])
    Push(word[constOff+1])
    return 0

  Push(GetHeapValue(gConstTable, tableIndex)) 
  result := 0  


PRI doGETFIELD | object, index, data
{{
/**
 * case jem#Jem_GETFIELD Get 16bit field from object.
 * Pop object/Push value
 * @returns 0 to let case know the function is implemented
 */
doGETFIELD();
}}
  index := Operand_UnsignedWord >> 1

  object := Pop
{
  if jazzDebug
    tv.Str(string($d,"doGETFIELD "))
    tv.Hex(object, 4)
    tv.Str(string(" index "))
    tv.Dec(index)
}
  data := GetObjectField(object, index)
{
  if jazzDebug
    tv.Str(string(" doGETFIELD value "))
    tv.Hex(data, 4 )
}
  Push(data)

  result := 0

    
PRI doPUTFIELD | object, index, value
{{
/**
 * case jem#Jem_PUTFIELD Set 16bit field in object.
 * Pop value/Pop object
 * @returns 0 to let case know the function is implemented
 */
doPUTFIELD();
}}
  index := Operand_UnSignedWord >> 1
  
  value := Pop
  object := Pop
{
  if jazzDebug
    tv.Str(string($d,"doPUTFIELD "))
    tv.Hex(object, 4)
    tv.Str(string(" index "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(value, 4 )
}
  SetObjectField(object, index, value)
    
  result := 0
  
PRI doNEWARRAY | arrayLen, arrayType
{{
/**
 *
 * case jem#Jem_NEWARRAY Create new array of operand type
 * NEWARRAY                                                                        
 * Meaning: Allocate new array.                                                    
 * Description: A new array of a specific array type, capable of holding size      
 * elements, is allocated. The result is a reference to the new object. Allocation 
 * of an array large enough to contain size items of the specific array type is    
 * attempted and all elements of the array are initialized to 0.                   
 * size represents the number of elements in the new array and must be an          
 * integer. The result is stored with an internal code that indicates the type of  
 * array to allocate. Possible values for the type of array are as follows:        
 * T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),        
 * T_INT(10), and T_LONG(11).                                                      
 * Note                                                                            
 * A NegativeArraySizeException is thrown if size is less than 0. An               
 * OutOfMemoryError is thrown if there is not enough memory to allocate the array. 
 *
 * @returns 0 to let case know the function is implemented
 */
doNEWARRAY();
}}
  arrayLen := Pop
  'Operand_UnsignedByte ' ***** HACK **** This is $00 so what's it mean ?
  arrayType := Operand_UnsignedByte  ' $00 means *ANY* arrayType
  
  case arrayType
    0: arrayType := jem#T_INT
    1: arrayType := jem#T_BYTE
    other: arrayType := jem#T_ANY
{    
  if jazzDebug
    tv.Str(string($d,"NEWARRAY "))
    tv.Dec(arrayType)
    tv.Str(string(" "))
    tv.Hex(arrayLen, 4 )
    tv.Out(" ")  
}  
  Push( NewArrayRef( arrayType, arrayLen+1 ) )

  result := 0
  'debugMsDelay := -1

   
PRI doANEWARRAY | arrayLen, index, ptr, ii
{{
/**
 * case jem#Jem_NEWARRAY Create new array of operand type
 * ANEWARRAY                                                                      
 * Meaning: Allocate new array of objects.                                        
 * Description: A new array of the indicated class type and capable of holding    
 * size elements is allocated. The result is a reference to the new object.       
 * Allocation of an array large enough to contain size elements of the given class
 * type is attempted and all elements of the array are initialized to null.       
 * size represents the number of elements in the new array and must be an         
 * integer. byte1 and byte2 are used to construct an index into the constant pool 
 * of the current class. When the item at that index is resolved, the resulting   
 * entry must be a class.                                                         
 * The anewarray instruction is used to create a single-dimension array.          
 * Note                                                                           
 * A NegativeArraySizeException is thrown if size is less than 0. An              
 * OutOfMemoryError is thrown if there is not enough memory to allocate the array.
 * @returns 0 to let case know the function is implemented
 */                                                            
doANEWARRAY();
}}
  arrayLen := Pop

  ' index into the class constant pool ... not sure what it means for now.
  index := Operand_UnsignedWord
   
  ptr := NewArrayRef(jem#T_ANY, arrayLen)

  ' set all entries to be null
  repeat ii from 0 to arrayLen-1
    SetHeapValue(ptr, ii, 0)
    
  Push(ptr)
  
  result := 0

PRI doINVOKEVIRTUAL | methodAddr, methOff, method, ii, classid, ptr, type, len, data, offset, oldclass
{{
/**
 * Will call virtual method when developed.
 * @TODO: Make this work :)
 * @returns non-zero for now because it is not compeletely implemented.
 */
doINVOKEVIRTUAL();   
}}

  ' garbage for now
  offset := Operand_UnSignedByte >> 1
  offset--
  methOff := Operand_UnsignedByte
{
  if jazzDebug
    tv.Str(string($d,"InvokeVirtual ... moff $"))
    tv.Hex(methOff, 4 )
    tv.str(string(" off "))  
    tv.Hex(offset, 4 )
}
  'ShowStack
  'ShowObject(Peek(sp,-offset))

  ' this gClassId business can most likely be replaced by
  ' getting same info from object at top of the stack.
  ' if the object is not at tos, invokevirtual will fail anyway.
  '
  'classid := LookupMethodClassId(gMethodId)
  classid := GetHeapValue(Peek(sp,-offset), ObjectThisClass)
{ 
  if jazzDebug
    tv.Str(string($d,"InvokeVirtual ... moff $"))
    tv.Hex(methOff, 4 )
    tv.str(string(" off "))  
    tv.Hex(offset, 4 )
    tv.str(string(" cid "))  
    tv.Hex(classid, 4 )
}
  methOff+=bytecodeStart
  methodAddr := methOff+classid
  method := byte[methodAddr + 3] << 8
  method |= byte[methodAddr + 4]            
{
  if jazzDebug
    tv.str(string(" madd "))  
    tv.Hex(methodAddr-bytecodeStart, 4 )
    tv.str(string(" method "))  
    tv.Hex(method, 4 )
}      
  'check if methodAddr points to a method in our class
  'methodAddr := Operand_UnSignedWord
  'classid := LookupMethodClassId(methodAddr)

  if classid == $118
    return 0
    
  if classid
    CallMethod(methodAddr, false)
    Push(oldclass)
    return 0
  else
    Crashed(jcc#InvokeVirtual_method_not_found)

  result := 0


PRI doINSTANCEOF | objectref, index, classid, type, len, data
{{
/**
 * find out if objectref is an instance of a class
 * INSTANCEOF                                                                      
 * Meaning: Determine if an object is of given type and return result.             
 * Description: Determines whether the referenced object can be cast to be a       
 * reference to an object of the expected class. This instruction will overwrite   
 * the object reference with 1 if the object is an instance of the expected class  
 * or one of its superclasses. Otherwise, the object reference is overwritten by 0.
 * @returns 0 to let case know the function is implemented
 */
doINSTANCEOF();
}}
  index := Operand_UnSignedWord  
  objectref := Pop

  GetHeapInfo(objectref, @type, @len, @data)
  classid := GetHeapDataValue(data, type, len, ObjectThisClass)
  
  if index == classid
    Push(1)
  else
    Push(0)
     
  result := 0


PRI doJazzLib : rc | ii, heapPtr, heapData, blockAddr, tmpAddr
{{
/**
 * Function is called from "notimplemented" function
 * @returns run status 0, of ok other wise no-zero
 */
doJazzLib();
}}
  result := -1

  case opcode
     
    jem#Jem_NEW:
      result := doNEW

    jem#Jem_GETFIELD:
      result := doGETFIELD

    jem#Jem_PUTFIELD:
      result := doPUTFIELD

    jem#Jem_NEWARRAY:
      result := doNEWARRAY

    jem#Jem_ANEWARRAY:
      result := doANEWARRAY

    jem#Jem_INVOKEVIRTUAL:
      result := doInvokeVirtual

    jem#Jem_INSTANCEOF:
      result := doINSTANCEOF


PUB jvmStep 
{{
/**
 * This is a debug control command that allows JVM to run until a breakpoint is hit.
 * @returns zero if break happened normally.
 */
__ int jvmStep();
}}
  result := jvmSpinDebug(0,debugging,0)

PUB jvmSingleStepBytecode 
{{
/**
 * This is a debug control command that allows JVM to run one bytecode "instruction" (opcode/operand).
 * An instruction is of the form OPCODE+OPERANDS, i.e. "bipush 6" or bytecode sequence $08,$06.
 * Many instructions take information from the stack and/or manipulate stack contents.
 * The state of the stack must be reasonable before many instructions can be executed.
 * The method blocks until the instruction is complete or an error occured. 
 */
__ int jvmSingleStepBytecode();
}}
  result := jvmSpinDebug(1,debugging,0)

PUB jvmRun 
{{
/**
 * This is a debug control command that allows JVM to run the bytecode program without debug control. 
 */
__ void jvmRun();
}}
  'debugMsDelay := 0
  'JvmInit    
  '\JvmSpinLoop

PUB jvmStop 
{{
/**
 * This is a debug control command. 
 * Stop the JVM from executing the next instruction and abort the current one if possible.
 * Hmm. This wont work unless jvm is running in a COG ..... 
 */
__ void jvmStop();
}}
  gAbort := true


PUB getMaxJvmBreakpoints 
{{
/**
 * This is a debug control command that tells user maximum number of breakpoints allowed.
 * @returns maximum number of breakpoints. 
 */
__ int getMaxJvmBreakpoints();
}}
  result := BREAKPOINT_SIZE

PUB setJvmBrkPtByNumber(num, addr) 
{{
/**
 * set a breakpoint by number
 * @param num is breakpoint index
 * @param addr is pc address for breakpoing
 */
 _ void setJvmBrkPtByNumber(int num, int addr);
}}
  if num => 0 and num < BREAKPOINT_SIZE
    gBreakpointArray[num] := addr

PUB getJvmBrkPtByNumber(num) 
{{
/**
 * set a breakpoint by number
 * @param num is breakpoint index
 * @param addr is pc address for breakpoing
 */
 _ void getJvmBrkPtByNumber(int num);
}}
  if num => 0 and num < BREAKPOINT_SIZE
    return gBreakpointArray[num]
    
PUB skipJvmBreakPointOnce
{{
/**
 * Skip breakpoint once. required to get past breakpoint.
 */
__ void skipJvmBreakPointOnce(void);
}}
  gGetOverBreak := true
  
PUB isBreakPoint(bppc) | ii
{{
/**
 * find out if PC is on a breakpoint
 * @param pc the program counter value to check for breakpoint
 * @returns non-zero or break number if pc is a breakpoint
 */
__ bool isBreakPoint(pc);
}}
  if bppc == 0
    return 0
    
  result := 0
  if gGetOverBreak
    gGetOverBreak := 0
    return false
  repeat ii from 0 to BREAKPOINT_SIZE-1
    if gBreakpointArray[ii] == bppc
      return ii+1


PUB getJvmAbortPointer 
{{
/**
 * This is command lets caller get the abort pointer.
 * This variable would be used to cause an execution abort for debugging.
 * @returns address of the JVM engine's abort variable.
 */
__ int getJvmAbortPointer();
}}
  result := @gAbort
  
PUB resetJvmPC
{{
/**
 * Sets PC to "main" program address.
 */
__ int resetJvmPC();
}}
  pc := gBytecodeMain+bytecodeStart

PUB getJvmPC 
{{
/**
 * This is a debug status command that returns the value of the bytecode offset program counter. 
 * This value will be the address in memory of the current PC instruction. 
 * @returns value of the program counter. 
 */
__ int getJvmPC();
}}
  result := pc
  
PUB getJvmSP 
{{
/**
 * This is a debug status command that returns the value of the stack pointer. 
 * This value will be the address in memory of the current SP location. 
 * @returns value of the stack pointer. 
 */
__ int getJvmSP();
}}
  result := sp

PUB getJvmSB 
{{
/**
 * This is a debug status command that returns the value of the stack base pointer. 
 * This value will be the address in memory of the current SB location. 
 * This value is a representation of the JVM's internal BP or base pointer. 
 * @returns value of the stack base pointer. 
 */
__ int getJvmSB();
}}
  result := bp
  
PUB getJvmFP 
{{
/**
 * This is a debug status command that returns the value of the Frame Pointer. 
 * The FP is defined as the beginning of the stack for the currently executing method. 
 * @returns value of the frame pointer. 
 */
__ int getJvmFP();
}}
  result := fp
  
PUB getJvmMP 
{{
/**
 * This is a debug status command that returns the relative bytecode offset value of the currently running 
 * Method's start address. This value is is the constant table method start address + the bytecode base address. 
 * @returns value of the method pointer. 
 */
__ int getJvmMP();
}}
  result := mb

PUB getJvmHP 
{{
/**
 * This is a debug status command that returns the value of the heap start pointer. 
 * @returns value of the heap start pointer. 
 */
__ int getJvmHP();
}}
  result := getJvmHeapStart
  
PUB getJvmHeapStart 
{{
/**
 * This is a debug status command that returns the current start position of the Heap pointer. 
 * @returns value of the heap pointer lowest memory position. 
 */
__ int getJvmHeapStart();
}}
  result := heap
  
PUB getJvmHeapEnd 
{{
/**
 * This is a debug status command that returns the current end position of the Heap pointer. 
 * @returns value of the heap pointer highest memory location.
 */
__ int getJvmHeapEnd();
}}
  result := HEAP_BASE

PUB getJvmLength 
{{
/**
 * Returns the length of bytecode
 * @returns bytecode length 
 */
__ int getJvmLength();
}}
  result := gBytecodeLen

PUB getJvmStart 
{{
/**
 * Returns the value of bytecodeStart address.
 * @returns bytecodeStart address 
 */
__ int getJvmStart();
}}
  result := bytecodeStart

PUB getJvmMain 
{{
/**
 * Returns the value of bytecode main address.
 * @returns main method address 
 */
__ int getJvmMain();
}}
  result := gBytecodeMain

PUB setJvmStart(addr)
{{
/**
 * sets the value of bytecodeStart address.
 * @param add is bytecodeStart address 
 */
__ int setJvmStart();
}}
  bytecodeStart := addr
  
PUB setJvmCallDebug(val)
{{
/**
 * This function sets callDebug state to value.
 * @param val if non-zero, print method call debug, else no call debug output 
 */
__ void setJvmCallDebug(int val);
}}
  callDebug := val

PUB setJvmVerboseDebug(val)
{{
/**
 * This function sets verbose debugging state to value.
 * @param val if non-zero, print verbose debug info, else no debug output 
 */
__ void setJvmVerboseDebug(int val);
}}
  debugging := val      