{{
**********************************
* jvmEngine - executes bytecodes *
**********************************
}}

CON
  _clkmode = xtal1+pll16x
  _xinfreq = 5_000_000

OBJ
  comm:  "jvmComm"    'communication port object
  jd:    "jvmData"    'jvm memory access object
  jvm:   "Jazzed_JVM_Engine_009"    
  jni:   "Jazzed_jvmNative"    'native functions
  'sio:   "Jazzed_SIO_Text"
  sio:   "Jazzed_VGA_Text"
  
DAT
  paddr long 0
  
PUB start
  sio.start(16)
  sio.str(string("jvmEngine ..."))
  comm.status(string("[JVM Engine] started"))
  'sio.start(31, 30, 0, 28800)
  jvm.start
     
PUB doBytecode: value | state,c
{{ Execute a bytecode }}

  state := jd.getActivity
  case state
    jd#kRun    : 'JVM running
                 c := doStep
                 if jd.mustNotifyStepDone
                   comm.notifyStepDone
                 return c
    jd#kBreak  : 'JVM reached breakpoint
                 'wait for IDE to send debug command (run/step/stop)
                 return -32768
    jd#kPassive: 'JVM was running but has stopped (exception,stopped,end)
                 return -32768
    jd#kJog    : 'appears not to be used
                 'so change state to kPassive
                 jd.setActivity(jd#kPassive)
                 return -32768 
    
  return -32768

PUB setBytecodeAddr(progaddr)

  ' if version is not one ignore address
  if byte[progaddr+2] <> 1
    return
{
  sio.str(string($d,"setBytecodeAddr "))
  sio.hex(progaddr, 4)
}    
  paddr := progaddr
  
  jvm.setJvmStart(progaddr)
  jvm.jvmInit

  'sio.str(string($d,"jvmInit done ...."))
  comm.status(string("[JVM Engine] program init done."))
            
PRI doStep | pc
{{ perform a step }}
  if paddr == 0
    return 0

  pc := jvm.getJvmPC
  sio.str(string($d,"doStep "))
  sio.hex(pc, 4)
  'comm.status(string("[JVM Engine] doStep"))
  
  if jvm.isBreakpoint(pc)
    sio.str(string("breakpoint "))
    comm.status(string("[JVM Engine] breakpoint"))    
    comm.breakpoint
    return 0

  result := jvm.doBytecode
  handleNative(result)

  'jvmDumpBreak
  
  if jvm.getJvmMP == 0 or jvm.getJvmFP == 0
    comm.end
    return -1
    
  result := 0

PRI handleNative(code) | stk, ii, rcnt, val

  ' native methods are positive integers
  if code < 0
    return
  'sio.str(string($d,"handleNative"))

  rcnt := jvm.Pop
  repeat while (rcnt-- > 0)
    val := jvm.GetVar(rcnt)
    result := jvm.GetHeapValuePtr(val,0)
    if result <> 0
      val := result
    jni.push(val)

  result := jni.do_native(code, @rcnt)
  if rcnt > 0
    jvm.Push(result)
   
PRI jvmDumpBreak | pc, opcode
  pc := jvm.getJvmPC
  opcode := byte[pc]
  if jvm.isBreakpoint(pc)
    sio.str(string($d,"Breakpoint PC $"))
    sio.hex(pc,4)
    sio.str(string(" JemPC $"))
    sio.hex(pc-jvm.getJvmStart,4)
    sio.out($d)
  jvm.JvmSpinDumpBreak(opcode, 0) 'ops.OpcodeNamePtr(opcode))
      